from ..ndtypes import Float64, make_array_type
from seq_expr import SeqExpr 

class ArrayExpr(SeqExpr):
  """
  Common base class for first-order array operations 
  that don't change the underlying data 
  """
  def __init__(self, array, type = None, source_info = None):
    self.array = array 
    self.type = type 
    self.source_info = source_info 
    
  def children(self):
    yield self.array 

class Array(ArrayExpr):
  def __init__(self, elts, type = None, source_info = None):
    self.elts = tuple(elts) 
    self.type = type 
    self.source_info = source_info 

  def __str__(self):
    return "[%s]" % ", ".join(str(e) for e in self.elts)
  
  def children(self):
    return self.elts

  def __hash__(self):
    return hash(self.elts)

class Slice(ArrayExpr):
  def __init__(self, start, stop, step, type = None, source_info = None):
    self.start = start 
    self.stop = stop 
    self.step = step 
    self.type = type 
    self.source_info = source_info 


  def __str__(self):
    return "%s:%s:%s"  % \
        (self.start.short_str(),
         self.stop.short_str(),
         self.step.short_str())

  def __repr__(self):
    return str(self)

  def children(self):
    yield self.start
    yield self.stop
    yield self.step

  def __eq__(self, other):
    return other.__class__ is Slice and \
           other.start == self.start and \
           other.stop == self.stop and \
           other.step == self.step
           
  def __hash__(self):
    return hash((self.start, self.stop, self.step))

class ConstArray(ArrayExpr):
  def __init__(self, shape, value, type = None, source_info = None):
    self.shape = shape 
    self.value = value 
    self.type = type 
    self.source_info = source_info 

  def children(self):
    yield self.shape 
    yield self.value 

class DiagonalArray(ArrayExpr):
  """
  Build a zero array with a filled value on its diagonal 
  Use this to implement np.diag, np.eye 
  """
  def __init__(self, shape, value, offset = None, type = None, source_info = None):
    self.shape = shape 
    self.value = value 
    self.offset = offset
    self.type = type 
    self.source_info = source_info
    
  def __str__(self):
    return "DiagonalArray(shape = %s, value = %s, offset = %s)" % \
      (self.shape, self.value, self.offset)
        
  def children(self):
    yield self.shape 
    yield self.value 
    if self.offset is not None:
      yield self.offset 

class ExtractDiagonal(ArrayExpr):
  """
  Go from an n-d input array to 1-d vector of diagonal elements
  """ 
  pass 

class ConstArrayLike(ArrayExpr):
  """
  Create an array with the same shape as the first arg, but with all values set
  to the second arg
  """

  def __init__(self, array, value, type = None, source_info = None):
    self.array = array 
    self.value = value 
    self.type = type 
    self.source_info = source_info 
  
  def children(self):
    yield self.array 
    yield self.value   

class Range(ArrayExpr):
  def __init__(self, start, stop, step, type = None, source_info = None):
    self.start = start 
    self.stop = stop 
    self.step = step 
    self.type = type 
    self.source_info = source_info 

  def children(self):
    yield self.start 
    yield self.stop 
    yield self.step 
  
  
  def __str__(self):
    return "Range(start = %s, stop = %s, step = %s)" % (self.start, self.stop, self.step)

class AllocArray(ArrayExpr):
  """Allocate an unfilled array of the given shape and type"""
  def __init__(self, shape, elt_type, type = None, source_info = None):
    # TODO: support a 'fill' field 
    self.shape = shape 
    self.elt_type = elt_type 
    self.type = type 
    self.source_info = source_info 

  def children(self):
    yield self.shape
    
  def __str__(self):
    return "AllocArray(shape = %s, elt_type = %s)" % (self.shape, self.elt_type)


class ArrayView(ArrayExpr):
  """Create a new view on already allocated underlying data"""
  def __init__(self, data, shape, strides, offset, size, type = None, source_info = None):
    self.data = data 
    self.shape = shape 
    self.strides = strides 
    self.offset = offset 
    self.size = size 
    self.type = type 
    self.source_info = source_info

  def __str__(self):
    return "ArrayView(data = %s, shape = %s, strides = %s, offset = %s, size = %s)" % \
      (self.data, self.shape, self.strides, self.offset, self.size)
      
  def children(self):
    yield self.data
    yield self.shape
    yield self.strides
    yield self.offset
    yield self.size

class Ravel(ArrayExpr):
  def children(self):
    return (self.array,)

  def __str__(self):
    return "Ravel(%s)" % self.array 

class Reshape(ArrayExpr):
  def __init__(self, array, shape, type = None, source_info = None):
    self.array = array 
    self.shape = shape 
    self.type = type 
    self.source_info = source_info

  
  def children(self):
    yield self.array 
    yield self.shape
    
  def __str__(self):
    return "Reshape(%s, %s)" % (self.array, self.shape)

class Shape(ArrayExpr):
  
  def __str__(self):
    return "Shape(%s)" % self.array 
  
class Strides(ArrayExpr):
  def __str__(self):
    return "Strides(%s)" % self.array 
  
    
class Transpose(ArrayExpr):
  def children(self):
    yield self.array
  
  def __str__(self):
    return "%s.T" % self.array 
  
class Tile(ArrayExpr):
  def __init__(self, array, reps, type = None, source_info = None):
    self.array = array 
    self.reps = reps 
    self.type = type 
    self.source_info = source_info 

  def children(self):
    yield self.array 
    yield self.reps 
    
  def __str__(self):
    return "Tile(%s, %s)" % (self.array, self.reps)
    




from dsltools.testing_helpers import eq
from run_function import specialize 
from ..transforms import Phase, Transform, CloneFunction
from ..transforms.pipeline import loopify

def transform_name(t):
  assert not isinstance(t, Phase)
  if hasattr(t, 'name') and t.name is not None:
    return t.name 
  elif hasattr(t, '__name__'):
    return t.__name__
  else:
    assert hasattr(t, '__class__')
    return t.__class__.__name__ 
  
def linearize_phase(typed_fn, phase):
  if isinstance(phase, (list, tuple)):
    return linearize_phases(typed_fn, phase)
  elif not isinstance(phase, Phase):
    name = transform_name(phase)
    return [ (name, phase) ]
  else:
    group_name = str(phase)
    previous = linearize_phase(typed_fn, phase.depends_on)
    if phase.run_if is not None and not phase.run_if(typed_fn):
      return previous  
    interleave = []
    for (name, t) in linearize_phases(typed_fn, phase.cleanup):
      interleave.append( ("  %s" % ( name, ), t) )
    combined = previous 
    for (name, t) in linearize_phases(typed_fn, phase.transforms):
      new_name = "%s:%s" % (group_name, name)
      combined.append((new_name, t))
      for (cleanup_name, t2) in interleave:
        cleanup_name = cleanup_name + " (after %s)" % new_name
        combined.append( (cleanup_name, t) )
    return combined 
    
def linearize_phases(typed_fn, phases):
  combined = []
  for phase in phases:
    combined.extend(linearize_phase(typed_fn, phase))
  return combined 
         
def get_transform_list(typed_fn, last_phase = loopify):
  return linearize_phase(typed_fn, last_phase)  

def find_broken_transform(fn, inputs, expected, 
                             print_transforms = True, 
                             print_functions = True, 
                             last_phase = loopify):
  from .. import interp 
  

  fn, args = specialize(fn, inputs, optimize=False)  

  try:  
    interp_result = interp.eval(fn, args)
  except:
    print "[Diagnose] Runtime before any optimizations while trying to run", fn
    raise  
    
  
   
  if not eq(interp_result, expected):
    print "[Diagnose] Expected:", expected 
    print "[Diagnose] Result:", interp_result
    print "[Diagnose] This function was busted before we optimized anything!" 
    return None 
  transforms = get_transform_list(fn, last_phase = last_phase)
  print "[Diagnose] Full list of transforms:"
  for (name, t) in transforms:
    print "  -> ", name 
  old_fn = fn 

  for (name, t) in transforms:
    print "[Diagnose] Running %s..." % (name, )
    # in case t is just a class, instantiate it 
    if not isinstance(t, Transform):
      t = t()
    assert isinstance(t, Transform)
    cloner = CloneFunction(parent_transform = t)
    new_fn = t.apply(cloner.apply(old_fn))
    if print_functions:
      print new_fn 
      
    result = interp.eval(new_fn, args)
    if not eq(result, expected):
      print "[Diagnose] Expected %s but got %s " % (expected, result)
      print "[Diagnose] After running ", t
      print "[Diagnose] Old function ", old_fn 
      print "[Diagnose] Transformed function ", new_fn 
      print "[Diagnose] The culprit was:", name 
      return t 
    old_fn = new_fn 
  print "[Diagnose] No problems found!"
  



from ..ndtypes import (UInt8, UInt16, UInt32, UInt64, 
                       Int8, Int16, Int32, Int64, Float32, Float64, NoneType,  
                       Bool, ArrayT, ClosureT, TupleT, SliceT, NoneT, PtrT)

_dtype_mappings = {
  Bool : "NPY_BOOL",
  Int8 : "NPY_INT8", 
  Int16 : "NPY_INT16",
  Int32 : "NPY_INT32",
  Int64 : "NPY_INT64",
  UInt8 : "NPY_UINT8",
  UInt16 : "NPY_UINT16",
  UInt32 : "NPY_UINT32",
  UInt64 : "NPY_UINT64",
  Float32 : "NPY_FLOAT32",
  Float64 : "NPY_FLOAT64"
}
def to_dtype(t):
  if t in _dtype_mappings:
    return _dtype_mappings[t]
  assert False, "Unsupported element type %s" % t
  
_ctype_mappings = {
  Int8:  "int8_t",
  UInt8:  "uint8_t",
  UInt16:  "uint16_t",
  UInt32:  "uint32_t",
  UInt64:  "uint64_t",
  Int16:  "int16_t",
  Int32:  "int32_t",
  Int64:  "int64_t",
  Float32:  "float",
  Float64:  "double",
  NoneType:  "int",
  Bool:  "int8_t",
}

def to_ctype(t):
  if t in _ctype_mappings:
    return _ctype_mappings[t]
  elif isinstance(t, PtrT):
    return "%s*" % to_ctype(t.elt_type)
  elif isinstance(t, (ArrayT, ClosureT, TupleT, SliceT, NoneT)):
    return "PyObject*"
  else:
    assert False, "Unsupported type %s" % t
    



import cStringIO
import os  
import sys 
import time
import tempfile 

class timer(object):
  def __init__(self, name = None, newline = True, 
               suppress_stdout = True, 
               suppress_stderr = True,
               propagate_exceptions = False):
    self.name = name 
    self.start_t = time.time()
    self.newline = newline
    self.suppress_stdout = suppress_stdout
    self.suppress_stderr = suppress_stderr
    self.propagate_exceptions = propagate_exceptions 
  def __enter__(self):
    if self.suppress_stdout:
      stdout_newfile = tempfile.NamedTemporaryFile()
      self.prev_stdout_fd = os.dup(sys.stdout.fileno())
      os.dup2(stdout_newfile.fileno(), sys.stdout.fileno())
      self.prev_stdout = sys.stdout
  
    if self.suppress_stderr:
      stderr_newfile = tempfile.NamedTemporaryFile()
      self.prev_stderr_fd = os.dup(sys.stderr.fileno())
      os.dup2(stderr_newfile.fileno(), sys.stderr.fileno())
      self.prev_stderr = sys.stderr 
    
    self.start_t = time.time()

  
  def elapsed(self):
    return time.time() - self.start_t
  
  def __exit__(self, exc_type, exc_value, traceback):
    t = self.elapsed()
    if self.suppress_stdout:
      os.dup2(self.prev_stdout_fd, self.prev_stdout.fileno())
    if self.suppress_stderr:
      os.dup2(self.prev_stderr_fd, self.prev_stderr.fileno())
    if self.newline:
      print 
    s = "Elapsed time: " if self.name is None else "%s : " % self.name 
    if exc_type is None:
      s += "%0.4f" % t
    else:
      name = str(exc_type) if exc_type.__name__ is None else exc_type.__name__
      s += "FAILED with %s '%s'" % (name, exc_value)
    print s 
    # don't raise exceptions
    if self.propagate_exceptions:
      return False 
    else:
      return exc_type is not KeyboardInterrupt
    




 
from .. syntax import Var, Tuple 
from syntax_visitor import SyntaxVisitor 

class FindLiveVars(SyntaxVisitor):
  def __init__(self):
    SyntaxVisitor.__init__(self)
    self.live_vars = set([])
    
  def visit_Var(self, expr):
    self.live_vars.add(expr.name)
    
  def visit_lhs(self, expr):
    expr_class = expr.__class__ 
    if expr_class is Var:
      pass 
    elif expr_class is Tuple:
      for elt in expr.elts:
        self.visit_lhs(elt)
    else:
      self.visit_expr(expr)
      
  def visit_fn(self, fn):
    self.live_vars.clear()
    for name in fn.arg_names:
      self.live_vars.add(name)
    self.visit_block(fn.body)
    return self.live_vars

def find_live_vars(fn):
  return FindLiveVars().visit_fn(fn)

class VarUseCount(SyntaxVisitor):
  
  def __init__(self):
    
    SyntaxVisitor.__init__(self)
    self.counts = {}
    
  def visit_Var(self, expr):
    old_count = self.counts.get(expr.name, 0)
    self.counts[expr.name] = old_count + 1 
  
  def visit_ForLoop(self, stmt):
    self.visit_Var(stmt.var)
    SyntaxVisitor.visit_ForLoop(self, stmt)
  
  def visit_lhs(self, expr):
    expr_class = expr.__class__ 
    if expr_class is Var:
      pass 
    elif expr_class is Tuple:
      for elt in expr.elts:
        self.visit_lhs(elt)
    else:
      self.visit_expr(expr)
      
  def visit_fn(self, fn):
    self.counts.clear()
    for name in fn.arg_names:
      self.counts[name] = 1
    self.visit_block(fn.body)
    return self.counts 

def use_count(fn):
  return VarUseCount().visit_fn(fn)



import llvm.core as core
import llvm.ee as ee
import llvm.passes as passes

import llvm_config 

class LLVM_Context:
  """Combine a module, exec engine, and pass manager into a single object"""

  _verify_passes = [
    'preverify',
    'domtree',
    'verify'
  ]

  _opt_passes = [
    'mem2reg',
    'targetlibinfo',
    'no-aa',
    'basicaa',
    'memdep',
    'tbaa',
    'instcombine',
    'simplifycfg',
    'basiccg',
    'memdep',
    'scalarrepl-ssa',
    'sroa',
    'domtree',
    'early-cse',
    'simplify-libcalls',
    'lazy-value-info',
    'correlated-propagation',
    'simplifycfg',
    'instcombine',
    'reassociate',
    'domtree',
    'mem2reg',
    'scev-aa',
    'loops',
    'loop-simplify',
    'lcssa',
    'loop-rotate',
    'licm',
    'lcssa',
    'loop-unswitch',
    'instcombine',
    'scalar-evolution',
    'loop-simplify',
    'lcssa',
    'indvars',
    'loop-idiom',
    'loop-deletion',
    #'loop-vectorize',
     'loop-unroll',
    #'bb-vectorize',
    'memdep',
    'gvn',
    'memdep',
    'sccp',
    'dse',
    'adce',
    'correlated-propagation',
    'jump-threading',
    'simplifycfg',
    'instcombine',
    'licm',
  ]

  def __init__(self, module_name, 
               optimize = llvm_config.llvm_optimize,
               verify = llvm_config.llvm_verify):
    self.module = core.Module.new(module_name)
    self.engine_builder = ee.EngineBuilder.new(self.module)
    self.engine_builder.force_jit()
    opt_level = 3 if optimize else 0
    if optimize:
      self.engine_builder.opt(opt_level)
    else:
      self.engine_builder.opt(opt_level)
    self.exec_engine = self.engine_builder.create()
    tm = ee.TargetMachine.new(opt = opt_level, cm=ee.CM_JITDEFAULT)
    self.tm = tm 
    _, fpm = passes.build_pass_managers(tm, 
                                     opt = opt_level,
                                     loop_vectorize = (opt_level > 0), 
                                     mod = self.module, 
                                     vectorize = (opt_level > 0), 
                                     )
    self.pass_manager = fpm 
    # self.fpm = fpm 
    for p in self._verify_passes:
      self.pass_manager.add(p)
    if optimize:
      for p in (self._opt_passes + self._verify_passes):
        self.pass_manager.add(p)

  def run_passes(self, llvm_fn, n_iters = llvm_config.llvm_num_passes):
    for _ in xrange(n_iters):
      self.pass_manager.run(llvm_fn)
    #passmanagers = passes.build_pass_managers(
    #        tm= self.tm , opt=3, inline_threshold=1000,
    #        loop_vectorize=True, vectorize=True, fpm=False)
    #for optname in self._opt_passes:
    #  passmanagers.pm.add(optname)
    #passmanagers.pm.add('loop-vectorize')
    #passmanagers.pm.run(self.module)
      
global_context = LLVM_Context("module")



import parakeet
import numpy as np 
 
def matmult_high_level(X,Y):
  return np.array([[np.dot(x,y) for y in Y.T] for x in X])


n, d = 1200, 1200
m = 1200
dtype = 'float32'
X = np.random.randn(m,d).astype(dtype)
Y = np.random.randn(d,n).astype(dtype)

from compare_perf import compare_perf

extra = {'BLAS (np.dot)' : np.dot}
try:
  import pycuda
  import pycuda.gpuarray 
  import pycuda.autoinit

  import scikits.cuda.linalg
  import scikits.cuda.cublas 
  cublas_context = scikits.cuda.cublas.cublasCreate()
  def gpu_dot(X,Y):
    X_gpu = pycuda.gpuarray.to_gpu(X)
    Y_gpu = pycuda.gpuarray.to_gpu(Y)
    Z_gpu = scikits.cuda.linalg.dot(X_gpu, Y_gpu, 'N', 'N', handle = cublas_context)
    return Z_gpu.get()
  extra['cuBLAS'] = gpu_dot 
except:
  print "Failed to import PyCUDA + scikits.cuda" 

try:
  import numba 

  @numba.autojit 
  def matmult_loops(X,Y,Z):
    m, d = X.shape
    n = Y.shape[1]
    for i in xrange(m):
      for j in xrange(n):
        total = X[i,0] * Y[0,j] 
        for k in xrange(1,d):
          total += X[i,k] * Y[k,j]
        Z[i,j] = total 
  
  def call_numba(X,Y):
    Z = np.zeros((X.shape[0],Y.shape[1])).astype(dtype)
    matmult_loops(X,Y,Z)
    return Z 

  extra['numba'] = call_numba 

except:
  print "Failed to import Numba" 
  pass 

compare_perf(matmult_high_level, [X,Y],
             cpython=True, 
             # numba can't run the nested comprehensions so we use
             # a special loopy version instead 
             numba=False,
             extra = extra, 
             suppress_output = False,
             propagate_exceptions = False)




import parakeet 
from parakeet import jit, testing_helpers, c_backend, config 
import numpy as np

c_backend.print_function_source = True
c_backend.print_module_source = True 

@jit
def avg1d(x):
  return sum(x) / float(len(x))

def test_avg1d():
  x = np.random.randn(20)
  testing_helpers.eq(x.mean(), avg1d(x))

@jit
def winmap_avg1d(x, w = 3):
  return parakeet.pmap1(avg1d, x, w)

def test_winmap_avg1d():
  x = np.random.randn(20)**2
  y = winmap_avg1d(x)
  assert x.shape == y.shape

 
@jit
def avg2d(x):
  nelts = x.shape[0] * x.shape[1]
  return sum(sum(x)) / float(nelts)

def test_avg2d():
  x = np.random.randn(20,30)
  testing_helpers.eq(x.mean(), avg2d(x))

@jit
def winmap_zeros(x, wx = 3, wy = 3):
  def zero(_):
    return 0
  return parakeet.pmap2(zero, x, (wx, wy))

def test_winmap_zeros():
  x = np.random.randn(100,100)
  y = winmap_zeros(x)
  assert y.sum() == 0

@jit
def winmap_first_elt(x, wx = 3, wy = 3):
  def f(window):
    return window[0,0]
  return parakeet.pmap2(f, x, wx, wy)


def test_window_shapes():
  x = np.array([0,1,2,3,4])
  def winshape(x):
    return x.shape[0] 
  y = parakeet.pmap1(winshape, x, 3)
  
  expected = np.array([2,3,3,3,2])
  assert y.shape == expected.shape, "Expected shape %s but got %s" % (expected.shape, y.shape)
  assert np.all(y == expected), "Expected array %s but got %s" % (expected, y)

@jit
def winavg2d( x, wx = 3, wy = 3):
  return parakeet.pmap2(avg2d, x, (wx, wy))

def test_winavg2d():
  x = np.random.rand(100,100)
  x[50:65, 50:65] = 0
  y = winavg2d(x)
  assert x.shape==y.shape
  assert x.max() >= y.max()
  assert x.min() <= y.min()
  
if __name__ == '__main__':
  testing_helpers.run_local_tests()
  



import numpy as np 

def fdtd(input_grid, steps):
    grid = input_grid.copy()
    old_grid = np.zeros_like(input_grid)
    previous_grid = np.zeros_like(input_grid)

    l_x = grid.shape[0]
    l_y = grid.shape[1]

    for i in range(steps):
        previous_grid[:, :] = old_grid
        old_grid[:, :] = grid 
        for x in range(l_x):
            for y in range(l_y):
                grid[x,y] = 0.0
                if x + 1 < l_x:
                    grid[x,y] += old_grid[x+1,y]
                if 0 < x-1 and x - 1 < l_x:
                    grid[x,y] += old_grid[x-1,y]
                if y+1 < l_y:
                    grid[x,y] += old_grid[x,y+1]
                if 0 < y-1 and y-1 < l_y:
                    grid[x,y] += old_grid[x,y-1]

                grid[x,y] /= 2.0
                grid[x,y] -= previous_grid[x,y]

    return grid

N = 1000
steps = 20 
input_grid = np.random.randn(N,N).astype('float32')

from compare_perf import compare_perf 
compare_perf(fdtd, [input_grid, steps], backends = ('c', 'openmp', 'cuda'))




import numpy as np 
import time 


import parakeet
from parakeet import testing_helpers 

def count_thresh_orig(values, thresh):
  n = 0
  for elt in values:
    n += elt < thresh
  return n

count_thresh = parakeet.jit(count_thresh_orig)

def np_thresh(values, thresh):
  return np.sum(values < thresh)

def par_thresh(values, thresh):
  return parakeet.sum(values < thresh)

def test_count_thresh():
  v = np.array([1.2, 1.4, 5.0, 2, 3])
  parakeet_result = count_thresh(v, 2.0)
  python_result = count_thresh_orig(v,2.0)
  testing_helpers.expect(count_thresh, [v, 2.0], count_thresh_orig(v, 2.0))


if __name__ == '__main__':
  testing_helpers.run_local_tests()




from .. import names 
from ..ndtypes import ClosureT, make_closure_type
from ..syntax import Expr, UntypedFn, Closure, Var, Return, ClosureElt, TypedFn, FormalArgs  
from ..syntax.helpers import get_types 

def unpack_closure(closure):
  """
  Given an object which could be either a function, a function's name, a
  closure, or a closure type:
  Return the underlying untyped function and the closure arguments
  """

  if closure.__class__ is ClosureT:
    fn, closure_args = closure.fn, closure.arg_types
  elif closure.__class__ is Closure:
    fn = closure.fn 
    closure_args = closure.args 
  elif closure.type.__class__ is ClosureT:
    fn, arg_types = closure.type.fn, closure.type.arg_types
    closure_args = \
        [ClosureElt(closure, i, type = arg_t)
         for (i, arg_t) in enumerate(arg_types)]
  else:
    fn = closure
    closure_args = []
    # fn = UntypedFn.registry[fn]
  return fn, closure_args


def make_typed_closure(clos, typed_fn):
  if clos.__class__ is UntypedFn:
    return typed_fn

  assert isinstance(clos, Expr) and clos.type.__class__ is ClosureT
  _, closure_args = unpack_closure(clos)
  if len(closure_args) == 0:
    return typed_fn
  else:
    t = make_closure_type(typed_fn, get_types(closure_args))
    return Closure(typed_fn, closure_args, t)

def mk_untyped_identity():
  var_name = names.fresh('x')
  var_expr = Var(var_name)
  fn_name = names.fresh('identity')
  args_obj = FormalArgs()
  args_obj.add_positional(var_name)
  return UntypedFn(name = fn_name, 
                   args = args_obj, 
                   body = [Return(var_expr)])

untyped_identity_function = mk_untyped_identity()

def _get_fundef(fn):
  c = fn.__class__ 
  if c is UntypedFn or c is TypedFn:
    return fn
  assert c is str, "Unexpected function %s : %s"  % (fn, fn.type)
  return UntypedFn.registry[fn]

def _get_closure_type(fn):
  c = fn.__class__ 
  if c is ClosureT:
    return fn
  elif c is Closure:
    return fn.type
  elif c is Var:
    assert fn.type.__class__ is ClosureT
    return fn.type
  else:
    fundef = _get_fundef(fn)
    return make_closure_type(fundef, [])



from ..frontend import jit, axis_macro 
from .. import prims 
from adverbs import scan 

from .. frontend import translate_function_value

from .. syntax import Reduce, Const 
from ..syntax.helpers import none, false, true, one_i32, zero_i32, zero_i24
 
from adverbs import reduce
from builtins import builtin_and, builtin_or
from parakeet.syntax.delay_until_typed import DelayUntilTyped

def _identity(x):
  return x 

def mk_reduce(combiner, x, init, axis):
  return Reduce(fn = translate_function_value(_identity), 
                combine = translate_function_value(combiner), 
                args = (x,),  
                init = init, 
                axis = axis)

@axis_macro
def reduce_min(x, axis = None):
  return mk_reduce(prims.minimum, x, init = None, axis = axis)


@axis_macro
def reduce_max(x, axis = None):
  return mk_reduce(prims.maximum, x, init = None, axis = axis)


@axis_macro 
def reduce_any(x, axis=None):
  return mk_reduce(builtin_or, x, init  = false, axis = axis, )

@axis_macro
def reduce_all(x, axis = None):
  return mk_reduce(builtin_and, x, init = true, axis = axis)

@axis_macro 
def reduce_sum(x, axis = None):
  return mk_reduce(prims.add, x, init = zero_i24, axis = axis)
  
@jit 
def builtin_min(x, y = None):
  if y is None:
    return reduce_min(x)
  else:
    return prims.minimum(x,y)

@jit
def builtin_max(x, y = None):
  if y is None:
    return reduce_max(x)
  else:
    return prims.maximum(x,y)


@jit 
def prod(x, axis=None):
  return reduce(prims.multiply, x, init=true, axis = axis)

@jit 
def mean(x, axis = None):
  return sum(x, axis = axis) / x.shape[0]

@jit 
def cumsum(x, axis = None):
  return scan(prims.add, x, axis = axis)

@jit 
def cumprod(x, axis = None):
  return scan(prims.multiply, x, axis = axis)

@jit 
def vdot(x,y):
  """
  Inner product between two 1Dvectors
  """
  return sum(x*y)

@jit
def argmax(x):
  """
  Currently assumes axis=None
  TODO: 
    - Support axis arguments
    - use IndexReduce instead of explicit loop
  
      def argmax_map(curr_idx):
        return curr_idx, x[curr_idx]
  
      def argmax_combine((i1,v1), (i2,v2)):
        if v1 > v2:
          return (i1,v1)
        else:
          return (i2,v2)
    
      return ireduce(combine=argmin_combine, shape=x.shape, map_fn=argmin_map, init = (0,x[0]))
  """
  bestval = x[0]
  bestidx = 0
  for i in xrange(1, len(x)):
    currval = x[i]
    if currval > bestval:
      bestval = currval
      bestidx = i
  return bestidx 

@jit
def argmin(x):
  """
  Currently assumes axis=None
  TODO: 
    - Support axis arguments
    - use IndexReduce instead of explicit loop
  
      def argmin_map(curr_idx):
        return curr_idx, x[curr_idx]
  
      def argmin_combine((i1,v1), (i2,v2)):
        if v1 < v2:
          return (i1,v1)
        else:
          return (i2,v2)
    
      return ireduce(combine=argmin_combine, shape=x.shape, map_fn=argmin_map, init = (0,x[0]))
  """
  bestval = x[0]
  bestidx = 0
  for i in xrange(1, len(x)):
    currval = x[i]
    if currval < bestval:
      bestval = currval
      bestidx = i
  return bestidx 





import itertools 


from .. import names 
from ..builder import build_fn 
from ..ndtypes import Int64, repeat_tuple, NoneType, ScalarT, TupleT, ArrayT 
from ..syntax import (ParFor, IndexReduce, IndexScan, Index, Map, OuterMap, Var, Const, Expr)
from ..syntax.helpers import get_types, none, zero_i64 
from ..syntax.adverb_helpers import max_rank_arg, max_rank 
from transform import Transform


class IndexifyAdverbs(Transform):
  """
  Take all the adverbs whose parameterizing functions assume they 
  get fed slices of input data and turn them into version which take explicit
  input indices
  """
  
  def fresh_input_name(self, expr):
    if expr is Var:
      return names.refresh(expr.name)
    else:
      return names.fresh("input")
  
  def fresh_fn_name(self, prefix, fn):
    return names.fresh(prefix + names.original(self.get_fn(fn).name))
  
  def fresh_index_names(self, n):
    name_supply = itertools.cycle(["i","j","k","l","ii","jj","kk","ll"])
    return [names.fresh(name_supply.next()) for _ in xrange(n)]
  
  _indexed_fn_cache = {}
  def indexify_fn(self, fn, 
                   axis,
                   array_args, 
                   cartesian_product = False,
                   output = None, 
                   index_offsets = None):
    """
    Take a function whose last k values are slices through input data 
    and transform it into a function which explicitly extracts its arguments
    """  
    array_args = tuple(array_args)
    
    array_arg_types = tuple(get_types(array_args))

    closure_args = self.closure_elts(fn)
    closure_arg_types = tuple(get_types(closure_args))
    n_closure_args = len(closure_args)
    fn = self.get_fn(fn)
    
    axes = self.normalize_axes(array_args, axis)
    
    
    key = (  fn.cache_key, 
             axes,
             closure_arg_types, 
             array_arg_types, 
             output is None,  
             cartesian_product, 
             index_offsets, 
           )

    def mk_closure():
      new_fn = self._indexed_fn_cache[key] 
      if output is None:
        return self.closure(new_fn, closure_args + array_args)
      else:
        return self.closure(new_fn, (output, ) + closure_args + array_args)
    
    if key in self._indexed_fn_cache:
      result = mk_closure()
      return result 
    
    if cartesian_product:
      # if we're doing a cartesian product then each argument may need 
      # a different number of indices depending on whether it's a scalar
      # and when its axis is None or given as an int 
      n_indices = 0
      for axis, arg_t in zip(axes, array_arg_types):
        if axis is None:
          n_indices += self.rank(arg_t)
        elif isinstance(arg_t, ArrayT):
          n_indices += 1
    else:
      # if we're doing an element-wise map, 
      # then either the axes are all None, in which case 
      # we need indices for the largest arg
      # or, we're just picking off one slice from 
      # every argument 
      if any(axis is None for axis in axes):
        assert all(axis is None for axis in axes), "Incompatible axes %s" % axes 
        n_indices = max_rank(array_arg_types)
      else:
        assert all(isinstance(axis, (int,long)) for axis in axes), "Invalid axes %s" % axes 
        n_indices = 1
        
    #index_input_type = Int64 if n_indices == 1 else repeat_tuple(Int64, n_arrays) 
    index_input_types = (Int64,) * n_indices  
    
    if output is None:
      inner_input_types = closure_arg_types + array_arg_types + index_input_types
      new_return_type = fn.return_type 
    else:
      inner_input_types = (output.type,) + closure_arg_types +  array_arg_types + index_input_types
      new_return_type = NoneType 
    
    input_names = []
    if output is not None:
      if output is Var:
        local_output_name = names.refresh(output.name)
      else:
        local_output_name = names.fresh("local_output")
      input_names.append(local_output_name) 
    
    for old_input_name in fn.arg_names:
      input_names.append(names.refresh(old_input_name)) 
    
    
    input_names.extend(self.fresh_index_names(n_indices))
      
    
    new_fn_name = self.fresh_fn_name("idx_", fn)
    
    new_fn, builder, input_vars = build_fn(inner_input_types, 
                                           new_return_type,
                                           name = new_fn_name,  
                                           input_names = input_names)

    index_input_vars = input_vars[-n_indices:]
    if output is None:
      output_var = None
      closure_arg_vars = input_vars[:n_closure_args]
      array_arg_vars = input_vars[n_closure_args:-n_indices]
    else:
      output_var = input_vars[0]
      closure_arg_vars = input_vars[1:n_closure_args+1]
      array_arg_vars = input_vars[n_closure_args+1:-n_indices]
    
    slice_values = \
      self.get_slices(builder, array_arg_vars, axes, index_input_vars, 
                      cartesian_product, index_offsets)


    elt_result = builder.call(fn, tuple(closure_arg_vars) + tuple(slice_values))
    if output is None: 
      builder.return_(elt_result)
    else:
      if len(index_input_vars) > 1:
        builder.setidx(output_var, builder.tuple(index_input_vars), elt_result)
      else:
        builder.setidx(output_var, index_input_vars[0], elt_result)
      builder.return_(none)
    
    new_fn.created_by = self.fn.created_by 
    new_fn.transform_history = self.fn.transform_history 
    self._indexed_fn_cache[key] = new_fn
    return mk_closure()
          
    
  
  def get_slices(self, builder, array_arg_vars, axes, index_input_vars, 
                 cartesian_product = False, index_offsets = None): 
    slice_values = []
    axes = self.normalize_axes(array_arg_vars, axes)
    # only gets incremented if we're doing a cartesian product
    if cartesian_product:
      idx_counter = 0
      for i, curr_array in enumerate(array_arg_vars):
        axis = axes[i]
        rank = curr_array.type.rank
        if rank <= 1 and axis is None:
          axis = 0
           
        if axis is None:
          start = idx_counter
          stop = idx_counter + rank
          curr_indices = index_input_vars[start:stop]
          idx_counter = stop 
          curr_slice = builder.index(curr_array, curr_indices)
        elif rank > axis:
          idx = index_input_vars[idx_counter]
          idx_counter += 1
          curr_slice = builder.slice_along_axis(curr_array, axis, idx)
        else:
          curr_slice = curr_array 
        slice_values.append(curr_slice)
    else:
      for i, curr_array in enumerate(array_arg_vars):
        axis = axes[i]
        rank = curr_array.type.rank

        if rank <= 1 and axis is None:
          axis = 0
        if axis is None:
          assert len(index_input_vars) <= rank, \
            "Insufficient indices for array arg %s : %s" % (curr_array, curr_array.type)
            
          # to be compatible with NumPy's broadcasting, we pull out the *last* r
          # indices so that Matrix + Vector will replicate the vector as columns, not rows
          curr_indices = index_input_vars[-rank:]
          curr_slice = builder.index(curr_array, curr_indices)
        elif rank > axis:
          curr_idx = index_input_vars[0]
          if index_offsets is not None:
            assert len(index_offsets) > i
            curr_offset = index_offsets[i]
            if not isinstance(curr_offset, Expr): 
              curr_offset = builder.int(curr_offset)
            curr_idx = builder.add(curr_idx, curr_offset)
          curr_slice = builder.slice_along_axis(curr_array, axis, curr_idx) 
        else:
          # if we're trying to map over axis 1 of a 1-d object, then there aren't
          # enough dims to slice anything, so it just gets passed in without modification
          
          curr_slice = curr_array 
        
          
        slice_values.append(curr_slice)
    return slice_values

  

  def create_map_output_array(self, 
                                 fn, array_args, axes, 
                                 cartesian_product = False, 
                                name = "output"):
    """
    Given a function and its argument, use shape inference to figure out the
    result shape of the array and preallocate it.  If the result should be a
    scalar, just return a scalar variable.
    """
    assert self.is_fn(fn), \
      "Expected function, got %s" % (fn,)
    assert isinstance(array_args, (list,tuple)), \
      "Expected list of array args, got %s" % (array_args,)
    axes = self.normalize_axes(array_args, axes)
    
    
    
    n_indices = 0
    for arg, axis in zip(array_args, axes):
      r = self.rank(arg)
      if r == 0:
        continue 
      if axis is None:
        if cartesian_product:
          n_indices += self.rank(arg)
        else:
          n_indices = max(n_indices, self.rank(arg))
      elif r <= axis:
        continue 
      else:
        if cartesian_product:
          n_indices += 1
        else:
          n_indices = max(n_indices, 1)
    
    
    # take the 0'th slice just to have a value in hand 
    inner_args = self.get_slices(builder = self, 
                                 array_arg_vars = array_args, 
                                 axes = axes, 
                                 index_input_vars = [zero_i64] * (n_indices),
                                 cartesian_product = cartesian_product)
     
                  
    
    if cartesian_product:
      extra_dims = []
      for array, axis in zip(array_args, axes):
        rank = self.rank(array)
        if axis is None:
          dim = 1
        elif rank > axis:
          dim = self.shape(array, axis)
        else:
          dim = 1 
        extra_dims.append(dim)
      outer_shape_tuple = self.tuple(extra_dims)
    else:
      outer_shape_tuple = self.iter_bounds(array_args, axes)
      if isinstance(outer_shape_tuple.type, ScalarT):
        outer_shape_tuple = self.tuple([outer_shape_tuple])

    return self.create_output_array(fn, inner_args, outer_shape_tuple, name)

  
  def transform_Map(self, expr, output = None):
    # TODO: 
    # - recursively descend down the function bodies to pull together nested ParFors
    

    args = self.transform_expr_list(expr.args)
    axes = self.normalize_axes(args, expr.axis)
    old_fn = expr.fn

    if output is None:
      output = self.create_map_output_array(old_fn, args, axes)

    bounds = self.iter_bounds(args, axes)
    index_fn = self.indexify_fn(expr.fn, axes, args, 
                                cartesian_product=False, 
                                output = output)
    self.parfor(index_fn, bounds)
    return output 
  
  def transform_OuterMap(self, expr):
    args = self.transform_expr_list(expr.args)
    axes = self.normalize_axes(args, expr.axis)
    
    fn = expr.fn 
    outer_shape = self.iter_bounds(args, axes, cartesian_product = True)
    # recursively descend down the function bodies to pull together nested ParFors 
    zero = self.int(0)
    
    first_values = [self.slice_along_axis(arg, axis, zero) 
                    for (arg,axis) in zip(args, axes)]
    # self.create_output_array(fn, inner_args, outer_shape, name)
    output =  self.create_output_array(fn, first_values, outer_shape)

    loop_body = self.indexify_fn(fn, axes, args, 
                                 cartesian_product = True, 
                                 output = output)
    self.parfor(loop_body, outer_shape)
    return output 
  
  def transform_IndexMap(self, expr, output = None):
    shape = expr.shape
    
    fn = expr.fn 
    dims = self.tuple_elts(shape)
    n_dims = len(dims)
    if n_dims == 1: shape = dims[0]
    if output is None: output = self.create_output_array(fn, [shape], shape)
    old_closure_args = self.closure_elts(fn)
    old_closure_arg_types = get_types(old_closure_args)
    fn = self.get_fn(fn)
    
    closure_arg_names = [self.fresh_input_name(clos_arg) for clos_arg in old_closure_args] 
    new_closure_vars = [Var(name, type=t) 
                        for name, t in 
                        zip(closure_arg_names, old_closure_arg_types)]
    
      
    old_input_types = fn.input_types
    last_input_type = old_input_types[-1]
    index_is_tuple = isinstance(last_input_type, TupleT)
    if index_is_tuple:
      index_types = last_input_type.elt_types
    else:
      index_types = old_input_types[-n_dims:]
    
    idx_names = self.fresh_index_names(n_dims)
    assert len(index_types) == n_dims, \
        "Mismatch between bounds of IndexMap %s and %d index formal arguments" % (dims, len(index_types))
    output_name = names.refresh("output")  
    
    new_input_names = [output_name] + closure_arg_names + idx_names            
    new_input_types =  [output.type]  + old_closure_arg_types + list(index_types)
    new_fn_name = names.fresh("idx_" + names.original(fn.name))
    new_fn, builder, input_vars = build_fn(new_input_types, NoneType,
                                           name = new_fn_name,  
                                           input_names = new_input_names)
    new_fn.created_by = self.fn.created_by 
    output_var = input_vars[0]
    
    idx_vars = input_vars[-n_dims:]
    
    if not self.is_none(expr.start_index):
      if isinstance(expr.start_index.type, ScalarT):
        idx_vars = [builder.add(idx, expr.start_index, "idx") for idx in idx_vars]
      else:
        start_indices = builder.tuple_elts(expr.start_index)
        assert len(start_indices) == len(idx_vars), \
          "Mismatch between number of indices %s and start offsets %s" % (idx_vars, start_indices)
        idx_vars = [builder.add(idx, offset, "idx%d") 
                    for i, (idx,offset) 
                    in enumerate(zip(idx_vars, start_indices))]
           
    
    if index_is_tuple:
      elt_result = builder.call(fn, new_closure_vars + [builder.tuple(idx_vars)])
    else:
      elt_result = builder.call(fn, new_closure_vars + idx_vars)
    if len(idx_vars) == 1:
      builder.setidx(output_var, idx_vars[0], elt_result)
    else:
      builder.setidx(output_var, builder.tuple(idx_vars), elt_result)

      
    builder.return_(none)
    new_closure = self.closure(new_fn, (output,) + tuple(old_closure_args)  )
    self.parfor(new_closure, shape)
    return output
  
  def transform_Reduce(self, expr):
    fn = expr.fn 
    combine = expr.combine 
    init = expr.init 
    
    args = []
    axes = []
    raw_axes = self.normalize_axes(expr.args, expr.axis)
    for axis, arg in zip(raw_axes, expr.args):
      if self.is_none(axis):
        args.append(self.ravel(arg))
        axes.append(0)
      else:
        args.append(arg)
        axes.append(axis)
    
    max_arg = max_rank_arg(args)
    nelts = self.shape(max_arg, axis)
    
    if self.is_none(axis):
      nelts = self.prod(nelts)
      
    if init is None or self.is_none(init):
      init_args = [self.index_along_axis(arg, axis, self.int(0)) for arg, axis in zip(args, axes)]
      init = self.call(fn, init_args)
      index_offsets = (1,) * len(axes)
      assert init.type == fn.return_type
      nelts = self.sub(nelts, self.int(1), "nelts") 
    else:
      index_offsets = None
    
    index_fn = self.indexify_fn(fn, 
                                axis, 
                                args, 
                                cartesian_product=False, 
                                index_offsets = index_offsets)

    return IndexReduce(fn = index_fn, 
                       init = init, 
                       combine = combine, 
                       shape = nelts, 
                       type = expr.type)
   
  def transform_Scan(self, expr, output = None):
    
    combine = expr.combine 
    init = expr.init 
    
    
    args = []
    axes = []
    for (arg, axis)  in zip(expr.args, self.normalize_axes(expr.args, expr.axis)):
      if self.is_none(axis):
        args.append(self.ravel(arg))
        axis = 0
      else:
        args.append(arg)
      axes.append(axis)
     
    bounds = self.iter_bounds(args, axes)

    if self.is_none(init):
      assert len(args) == 1
      init_args = [self.index_along_axis(arg, axis, self.int(0))
                   for arg in args]
      init = self.call(expr.fn, init_args)  
      index_offsets = (1,) * len(bounds)
      bounds = tuple(self.sub(bound, self.int(1), "scan_niters") for bound in bounds)
    else:
      index_offsets = None
      
    index_fn = self.indexify_fn(expr.fn, 
                                axis, 
                                args, 
                                cartesian_product=False, 
                                index_offsets = index_offsets)
    

    return IndexScan(fn = index_fn,
                     init = init, 
                     combine = combine,
                     emit = expr.emit, 
                     shape = bounds,
                     type = expr.type)
  
  def transform_Filter(self, expr):
    assert False, "Filter not implemented"
    
  
  def transform_Assign(self, stmt):
    """
    If you encounter an adverb being written to an output location, 
    then why not just use that as the output directly? 
    """
    if stmt.lhs.__class__ is Index:
      rhs_class = stmt.rhs.__class__ 
      if rhs_class is Map:
        self.transform_Map(stmt.rhs, output = stmt.lhs)
        return None 
      elif rhs_class is OuterMap:
        self.transform_OuterMap(stmt.rhs, output = stmt.lhs)
    return Transform.transform_Assign(self, stmt)

  



from .. ndtypes import ArrayT, NoneT, NoneType, ScalarT, ClosureT, TupleT, FnT, Type, SliceT, PtrT
from .. ndtypes import lower_rank 

from .. syntax import Expr, Tuple, Var, Index, Closure, TypedFn 
from .. syntax.helpers import get_types, get_elt_types 
from collect_vars import collect_binding_names
from syntax_visitor import SyntaxVisitor

class Verify(SyntaxVisitor):
  call_stack = []
  def __init__(self, fn):
    SyntaxVisitor.__init__(self)
    
    self.fn = fn
    self.call_stack.append(fn)
    self.bound = set(fn.arg_names)
    self.seen_return = False 

  def __del__(self):
    self.call_stack.pop()
    
  def bind_var(self, name):
    assert name not in self.bound, \
        "Error: variable %s already bound" % name
    self.bound.add(name)
    assert name in self.fn.type_env, \
        "Assigned variable %s has no entry in type dictionary" % name

  def phi_value_ok(self, lhs_name, v):
    assert isinstance(v, Expr), \
       "Invalid expression in phi node: %s" % (v,)
    self.visit_expr(v)
    assert v.type is not None, \
       "Error in phi node: value %s has no type annotation" % v
    t = self.fn.type_env[lhs_name]
    assert v.type == t, \
       "Invalid type annotation on %v, expected %s but got %s" % (v,t,v.type)

  def visit_merge_loop_start(self, phi_nodes):
    for (k, (v,_)) in phi_nodes.iteritems():
      self.bind_var(k)
      self.phi_value_ok(k,v)

  def visit_merge_loop_repeat(self, phi_nodes):
    for (k, (left_value, right_value)) in phi_nodes.iteritems():
      self.phi_value_ok(k, left_value)
      self.phi_value_ok(k, right_value)

  def visit_merge_if(self, phi_nodes):
    for (k, (left_value, right_value)) in phi_nodes.iteritems():
      self.bind_var(k)
      self.phi_value_ok(k, left_value)
      self.phi_value_ok(k, right_value)

  def visit_Var(self, expr):
    assert expr.name in self.bound, \
        "Variable %s used before assignment" % expr.name
    assert expr.name in self.fn.type_env, \
        "Variable %s has no entry in the type dictionary" % expr.name
    assert expr.type == self.fn.type_env[expr.name], \
        "Variable %s in fn %s should have type %s but annotated with %s" % \
        (expr.name, self.fn, self.fn.type_env[expr.name], expr.type)

  def visit_PrimCall(self, expr):
    self.visit_expr_list(expr.args)

    for arg in expr.args:
      assert arg.type is not None, \
          "Expected type annotation for %s" % (arg, )
      assert isinstance(arg.type, ScalarT), \
          "Can't call primitive %s with argument %s of non-scalar type %s" % \
          (expr.prim, arg, arg.type)

  def visit_expr(self, expr):
    assert expr is not None, "Expression missing, must be a compiler bug"
    assert expr.type is not None, \
      "Missing type annotation on %s" % expr 
    SyntaxVisitor.visit_expr(self, expr)

  def visit_Const(self, expr):
    assert isinstance(expr.type, (ScalarT, NoneT)), \
      "Invalid constant %s : %s" % (expr.value, expr.type)
       
  def visit_ExprStmt(self, stmt):
    self.visit_expr(stmt.value)
    assert stmt.value.type and stmt.value.type.__class__ is NoneT, \
      "Expected effectful expression %s to have type %s but instead got %s" % \
      (stmt.value, NoneType, stmt.value.type)
  
  def visit_Index(self, expr):
    arr = expr.value
    arr_t = arr.type  
    idx = expr.index 
    idx_t = idx.type 
    assert arr.type is not None, \
      "Error in indexing expression %s: %s lacks type" % (expr, arr) 
    assert idx.type is not None, \
      "Error in indexing expression %s: %s lacks type" % (expr, idx)
    assert isinstance(arr.type, (TupleT, ArrayT, PtrT)), \
      "Expected %s to be an array, pointer, or tuple" % arr
    assert isinstance(idx.type, (TupleT, SliceT, NoneT, ScalarT)), \
      "Expected index %s to be scalar or tuple" % idx
     
    if isinstance(arr_t, ArrayT):
      rank  = arr_t.rank
      if isinstance(idx_t, TupleT):
        n_indices = len(idx_t.elt_types)
      else: 
        n_indices = 1 
      assert rank >= n_indices, \
        "Expected at most %d indices but found %d in expression %s" % (rank, n_indices, expr)
  
  def check_fn_args(self, fn, args = None, arg_types = None):
    if arg_types is None: 
      assert args is not None, "Function args missing" 
      arg_types = get_types(args)
    n_given = len(arg_types)
    n_expected = len(fn.input_types)
    assert n_given == n_expected, \
        "Arity mismatch when calling %s, expected %d but got %d" % \
        (fn.name, n_expected, n_given)
    for (i, arg_name) in enumerate(fn.arg_names):
      given_t = arg_types[i]
      expected_t = fn.type_env[arg_name]
      signature_t = fn.input_types[i] 
      assert expected_t == signature_t, \
          "Function %s has inconsistent types %s and %s for arg %s" % \
          (fn.name, expected_t, signature_t, arg_name)
      assert given_t == expected_t, \
          "Given argument %s%s doesn't matched expected '%s' : %s in %s" % \
          ("'%s' : " %  args[i] if args is not None else "", 
           given_t, 
           arg_name, 
           expected_t, 
           fn.name)  
  
  
  def get_fn_and_closure(self, fn):
    if fn.__class__ is Closure: 
      return fn.fn, tuple(get_types(fn.args)) 
    elif fn.__class__ is TypedFn:
      return fn, ()
    elif fn.type.__class__ is FnT:
      return  fn, ()
    else:
      assert isinstance(fn.type, ClosureT), "Unexpected function %s : %s" % (fn, fn.type)
  
  
  def verify_call(self, fn, args):
    fn, closure_arg_types = self.get_fn_and_closure(fn)
    arg_types = []
    for arg in args:
      if isinstance(arg, Type):
        arg_types.append(arg)
      else:
        assert isinstance(arg, Expr), \
          "Expected call argument to be expression but got %s" % arg 
        arg_types.append(arg.type)
    arg_types = tuple(closure_arg_types) + tuple(arg_types)
    try:
      self.check_fn_args(fn, arg_types = arg_types)
    except:
      print "[verify] Errors in function call %s(%s)" % (fn.name, args) 
      raise 
  
      
  def visit_Call(self, expr):
    self.verify_call(expr.fn, expr.args)
  
  def is_none(self, expr):
    return expr is None or (hasattr(expr, 'type') and isinstance(expr.type, NoneT))
  
  def visit_Map(self, expr):
    assert len(expr.args) > 0, "Can't have Map without array arguments"
    if expr.fn.__class__ is Closure: 
      closure_elts = tuple(expr.fn.args) 
      fn = expr.fn.fn 
    else:
      fn = expr.fn
      closure_elts = ()
    verify(fn)
    if self.is_none(expr.axis):
      elt_types = get_elt_types(expr.args)
    else:
      elt_types = [lower_rank(arg.type, 1) for arg in expr.args]
    arg_types = tuple(get_types(closure_elts)) + tuple(elt_types)
    args = tuple(closure_elts) + tuple(expr.args)                      
    self.check_fn_args(fn, args, arg_types)
  
  def visit_Return(self, stmt):
    self.visit_expr(stmt.value)
    assert stmt.value.type is not None and stmt.value.type == self.fn.return_type, \
        "Incorrect return type in %s: ret value %s, expected %s but got %s" % \
        (self.fn.name, stmt.value, self.fn.return_type, stmt.value.type)
    self.seen_return = True 
    
  def visit_lhs(self, lhs):
    c = lhs.__class__
    assert c in (Var, Tuple, Index), "Invalid LHS of assignment: %s" % lhs 
    if c is Tuple:
      for elt in lhs.elts:
        self.visit_lhs(elt)

  def visit_Assign(self, stmt):
    assert stmt.lhs.type is not None, \
        "Missing LHS type for assignment %s" % stmt
    assert stmt.rhs.type is not None, \
        "Missing RHS type for assignment %s" % stmt
    self.visit_lhs(stmt.lhs)
    lhs_names = collect_binding_names(stmt.lhs)
    for lhs_name in lhs_names:
      self.bind_var(lhs_name)
    self.visit_expr(stmt.rhs)
    rhs_t = stmt.rhs.type 
    if stmt.lhs.__class__ is Index:
      array_t = stmt.lhs.value.type 
      assert stmt.lhs.type == rhs_t or rhs_t == array_t.elt_type, \
        "Mismatch between LHS type %s and RHS %s in '%s'" % \
        (stmt.lhs.type, stmt.rhs.type, stmt)
    else:
      assert stmt.lhs.type == stmt.rhs.type, \
        "Mismatch between LHS type %s and RHS %s in '%s'" % \
        (stmt.lhs.type, stmt.rhs.type, stmt)

  def visit_ForLoop(self, stmt):
    assert stmt.var.__class__ is Var, "Expected loop variable, got: %s" % stmt.var 
    self.bind_var(stmt.var.name)
    self.visit_expr(stmt.var)

    assert stmt.start.type == stmt.var.type, \
      "Loop variable type %s : %s doesn't match start value %s : %s" % \
      (stmt.var, stmt.var.type, stmt.start, stmt.start.type)
    self.visit_expr(stmt.start)
    self.visit_merge_loop_start(stmt.merge)

    assert stmt.stop.type == stmt.var.type, \
      "Loop variable type %s : %s doesn't match stop value %s : %s" % \
      (stmt.var, stmt.var.type, stmt.stop, stmt.stop.type)
    self.visit_expr(stmt.stop)
    self.visit_block(stmt.body)
    
    assert stmt.step.type == stmt.var.type, \
     "Loop variable type %s : %s doesn't match stop value %s : %s" % \
      (stmt.var, stmt.var.type, stmt.stop, stmt.stop.type)
    self.visit_expr(stmt.step)
    self.visit_merge_loop_repeat(stmt.merge)
    
  def visit_ParFor(self, stmt):
    """
    Have to be annoyingly flexible with ParFor since we might encounter
    ordinary or flattened functions as arguments and the latter don't 
    take tuples of indices but have to take each index as a separate parameter
    """
    self.visit_expr(stmt.bounds)
    fn, closure_arg_types = self.get_fn_and_closure(stmt.fn)
    verify(fn)
    bounds_t = stmt.bounds.type
    assert len(fn.input_types) > 0, \
      "Problem with ParFor, can't call function %s which accepts 0 args" % fn.name 
       
    last_input_type = fn.input_types[-1]
    if isinstance(last_input_type, TupleT):
      assert bounds_t == last_input_type, \
        "ParFor function expects index type %s but bounds of type %s" % (last_input_type, bounds_t)
      arg_types = (bounds_t,)
    elif isinstance(bounds_t, TupleT):
      arg_types = bounds_t.elt_types
    else:
      arg_types = (stmt.bounds.type,)
    self.verify_call(fn, tuple(closure_arg_types) + tuple(arg_types))
    
  def visit_stmt(self, stmt):
    assert stmt is not None, "Statement missing, must be a compiler bug"
    SyntaxVisitor.visit_stmt(self, stmt)

  def visit_TypedFn(self, fn):
    return verify(fn)

def verify(fn):
  n_input_types = len(fn.input_types)
  n_arg_names = len(fn.arg_names)
  assert n_input_types == n_arg_names, \
     "Function %s has %d input types and %d input names" % \
     (fn.name, n_input_types, n_arg_names)
  for arg_name, input_t in zip(fn.arg_names, fn.input_types):
    assert arg_name in fn.type_env, \
        "Input %s has no entry in %s's type environment" % \
        (arg_name, fn.name)
    t = fn.type_env[arg_name]
    assert input_t == t, \
        "Mismatch between %s's input type %s and the arg %s's type %s" % \
        (fn.name, input_t, arg_name, t)
  try:    
    verifier = Verify(fn) 
    verifier.visit_block(fn.body)
    assert verifier.seen_return, "Never encountered Return statement"
  except:
    if fn.name == Verify.call_stack[-1].name:
      print "Error in function"
      print "---------------------"
    else:
      print "...called from:"
    print fn
    print 
    raise 



__author__  = 'Alex Rubinsteyn'
__email__   = 'alex -dot- rubinsteyn -at- gmail -dot- com'
__desc__    = 'Runtime compiler for numerical Python'
__license__     = 'BSD3'
__version__     = '0.23.2'
__website__     = 'http://www.parakeetpython.com' 




from ..ndtypes import TupleT, Int64 
from ..syntax import Index, Map, unwrap_constant, zero_i64 
from transform import Transform
from parakeet.syntax.stmt import ForLoop

class LowerAdverbs(Transform):
    
  def transform_TypedFn(self, expr):
    from pipeline import loopify  
    return loopify(expr)
  

  
  def transform_IndexMap(self, expr, output = None):  
    # recursively descend down the function bodies to pull together nested ParFors
    fn = self.transform_expr(expr.fn)
    shape = expr.shape 
    
    dims = self.tuple_elts(shape)
    if len(dims) == 1:
      shape = dims[0]
    
    if output is None:
      output = self.create_output_array(fn, [shape], shape)

    def loop_body(idx):
      elt_result =  self.call(fn, (idx,))
      self.setidx(output, idx, elt_result)
    self.nested_loops(dims, loop_body)    
    return output
  
  _loop_counters = ["i", "j", "k", "l", "ii", "jj", "kk", "ll"]
  
  def get_loop_counter(self, depth):
    loop_counter_name = self._loop_counters[depth % len(self._loop_counters)] 
    return  self.fresh_var(Int64, loop_counter_name)
  
   
  def build_nested_reduction(self, indices, starts, bounds, old_acc, body_fn):
      if len(bounds) == 0:
        if len(indices) > 1:
          indices = self.tuple(indices)
        else:
          indices = indices[0]
        new_acc = body_fn(indices, old_acc)
        return new_acc 
      else:
        acc_t = old_acc.type 
        acc_before = self.fresh_var(acc_t, "acc_before")
        loop_counter = self.get_loop_counter(len(indices)) 
        
        future_indices = indices + (loop_counter,)
        future_starts = starts[1:]
        future_bounds = bounds[1:]
        self.blocks.push()
        acc_after = self.build_nested_reduction(future_indices, 
                                                future_starts, 
                                                future_bounds, 
                                                acc_before, 
                                                body_fn)
        body = self.blocks.pop()
        merge = {acc_before.name : (old_acc, acc_after)}
        for_loop = ForLoop(var = loop_counter, 
                           start = starts[0],
                           stop = bounds[0], 
                           step = self.int(1), 
                           body = body, 
                           merge = merge)
        self.blocks.append_to_current(for_loop)
        return acc_before # should this be acc_after?

  
  def transform_IndexReduce(self, expr):
    init = self.transform_if_expr(expr.init)
    fn = self.transform_expr(expr.fn)
    combine = self.transform_expr(expr.combine)
    shape = expr.shape 
    # n_loops = len(dims)
    
    assert init is not None, "Can't have empty 'init' field for IndexReduce"
   
    if isinstance(shape.type, TupleT): 
      bounds = self.tuple_elts(shape)
    else:
      bounds = [shape]
      
    if expr.start_index is not None:
      if isinstance(expr.start_index.type, TupleT):
        starts = self.tuple_elts(expr.start_index)
      else:
        starts = [expr.start_index]
    else:
      starts = [self.int(0)] * len(bounds)
      
    
    def body(indices, old_acc):
      elt = self.call(fn, (indices,))
      return self.call(combine, (old_acc, elt))
       
    return self.build_nested_reduction(
              indices = (), 
              starts = starts, 
              bounds = bounds, 
              old_acc = init, 
              body_fn = body)


  def transform_IndexScan(self, expr, output = None):
    init = self.transform_if_expr(expr.init)
    fn = self.transform_expr(expr.fn)
    combine = self.transform_expr(expr.combine)
    
    shape = expr.shape 
    if output is None:
      output = self.create_output_array(fn, [shape], shape)
      
    
    assert init is not None, "Can't have empty 'init' field for IndexScan"
    assert init.type == self.return_type(fn), \
      "Mismatching types init=%s, fn returns %s" % (init.type, self.return_type(fn))
    if isinstance(shape.type, TupleT): 
      bounds = self.tuple_elts(shape)
    else:
      bounds = [shape]
      
    def body(indices, old_acc):
      elt = self.call(fn, (indices,))
      new_acc = self.call(combine, (old_acc, elt))
      self.setidx(output, indices, new_acc)
      return new_acc
    starts = [self.int(0)] * len(bounds)
    self.build_nested_reduction(indices = (),
                                starts = starts, 
                                bounds = bounds, 
                                old_acc = init, body_fn = body)
    return output 
  
  def transform_IndexFilter(self, expr):
    assert False, "IndexFilter not implemented"

  
  def transform_IndexFilterReduce(self, expr):
    assert False, "IndexFilterReduce not implemented"
    
  

  



import numpy as np

import parakeet
from parakeet import config, each, syntax
from parakeet.transforms.pipeline import lowering, high_level_optimizations
from parakeet.analysis.syntax_visitor import SyntaxVisitor
from parakeet.testing_helpers import expect, run_local_tests


def A(x):
  return x + 1

def B(x):
  return A(x)

def C(x):
  return B(x)

def simple_args(exprs):
  return all(simple_expr(e) for e in exprs)

def simple_primcall(e):
  return isinstance(e, syntax.PrimCall) and simple_args(e.args)

def simple_expr(expr):
  return isinstance(expr, (syntax.Const, syntax.Var)) or \
         simple_primcall(expr)

def simple_stmt(stmt):
  """
  Is this a statement from a simple straightline function?

  (can only contain scalar computations and no control flow)
  """
  if isinstance(stmt, syntax.Return):
    return simple_expr(stmt.value)
  elif isinstance(stmt, syntax.Assign):
    return simple_expr(stmt.rhs)
  else:
    return False

def simple_block(stmts):
  return all(simple_stmt(s) for s in stmts)

def simple_fn(fn):
  return simple_block(fn.body)

def test_inline():
  typed_fn = parakeet.typed_repr(C, [1])
  assert len(typed_fn.body) in [1,2], \
      "Expected body to be 1 or 2 statements, got %s" % typed_fn
  assert simple_fn(typed_fn)
  expect(C, [1], 2)

def lots_of_arith(x):
  y = 4 * 1
  z = y + 1
  a = z / 5
  b = x * a
  return b

def test_simple_constant_folding():
  expect(lots_of_arith, [1], 1)
  typed_fn = parakeet.typed_repr(lots_of_arith, [1], optimize = True)
  assert len(typed_fn.body) == 1, \
      "Insufficiently simplified body: %s" % typed_fn

def const_across_control_flow(b):
  if b:
    x = 1
  else:
    x = 1
  return x

def test_constants_across_control_flow():
  expect(const_across_control_flow, [True], 1)
  typed_fn = parakeet.typed_repr(const_across_control_flow, [True])
  assert len(typed_fn.body) == 1, "Fn body too long: " + str(typed_fn.body)
  stmt = typed_fn.body[0]
  assert isinstance(stmt, syntax.Return)
  assert isinstance(stmt.value, syntax.Const)

def always_true_branch():
  x = 1 + 1
  if x == 2:
    res = 0 + 0
    return res
  else:
    res = 1 * 1 + 0
    return res

def test_always_true():
  expect(always_true_branch, [], 0)
  typed_fn = parakeet.typed_repr(always_true_branch, [])
  assert len(typed_fn.body) == 1, "Fn body too long: " + str(typed_fn.body)
  stmt = typed_fn.body[0]
  assert isinstance(stmt, syntax.Return)
  assert isinstance(stmt.value, syntax.Const)

def always_false_branch():
  x = 1 + 2
  if x == 2:
    res = 1 * 0 + 0
    return res
  else:
    res = 0 + 1 * 1
    return res

def test_always_false():
  expect(always_false_branch, [], 1)
  typed_fn = parakeet.typed_repr(always_false_branch, [])
  assert len(typed_fn.body) == 1, "Fn body too long: " + str(typed_fn.body)
  stmt = typed_fn.body[0]
  assert isinstance(stmt, syntax.Return)
  assert isinstance(stmt.value, syntax.Const)


def g(x):
  def h(xi):
    return xi + 1.0
  return each(h,x)

def nested_add1(X):
  return each(g, X)

class CountLoops(SyntaxVisitor):
  def __init__(self):
    SyntaxVisitor.__init__(self)
    self.count = 0

  def visit_While(self, stmt):
    self.count += 1
    SyntaxVisitor.visit_While(self, stmt)

def count_loops(fn):
  Counter = CountLoops()
  Counter.visit_fn(fn)
  return Counter.count

def test_copy_elimination():
  x = np.array([[1,2,3],[4,5,6]])
  expect(nested_add1, [x], x + 1.0)
  typed_fn = parakeet.typed_repr(nested_add1, [x])
  lowered = lowering.apply(typed_fn)
  n_loops = count_loops(lowered)
  n_expected = 3 if config.opt_loop_unrolling else 2
  assert n_loops <= n_expected, \
      "Too many loops generated! Expected at most 2, got %d" % n_loops

def allpairs_dist(x):
  return np.array([[np.sqrt(np.sum( (x1-x2)**2)) for x2 in x] for x1 in x])

from parakeet.syntax import OuterMap, Return  
def test_combine_nested_maps():
  x = np.array([[1,2], [3,4]])
  typed_fn = parakeet.typed_repr(allpairs_dist, [x])
  typed_fn = high_level_optimizations.apply(typed_fn)
  assert len(typed_fn.body) == 1
  stmt = typed_fn.body[0]
  assert stmt.__class__ is Return 
  v = stmt.value 
  assert v.__class__ is OuterMap
  assert len(v.args) == 2
  
if __name__ == '__main__':
  run_local_tests()




from .. builder import build_fn 
from .. ndtypes import (NoneT, ScalarT, Int64, SliceT, TupleT, NoneType,repeat_tuple)
from .. syntax import  Index, Tuple, Var, ArrayView
from ..syntax.helpers import zero_i64, one_i64, all_scalars, slice_none, none


from transform import Transform
from phase import Phase 


class LowerSlices(Transform):  
  

  
  _setidx_cache = {}
  def make_setidx_fn(self, lhs_array_type, 
                             rhs_value_type, 
                             fixed_positions, 
                             slice_positions):
    fixed_positions = tuple(fixed_positions)
    slice_positions = tuple(slice_positions)
    key = lhs_array_type, rhs_value_type, fixed_positions, slice_positions
    if key in self._setidx_cache: 
      return self._setidx_cache[key]

    
    n_fixed_indices = len(fixed_positions)
    n_parfor_indices = len(slice_positions)
    n_indices = n_fixed_indices + n_parfor_indices
    parfor_idx_t = repeat_tuple(Int64, n_parfor_indices) if n_parfor_indices > 1 else Int64 
    fixed_index_types = [Int64] * n_fixed_indices
    slice_start_types = [Int64] * n_parfor_indices
    slice_step_types = [Int64] * n_parfor_indices
    input_types = [lhs_array_type, rhs_value_type] + fixed_index_types + \
      slice_start_types + slice_step_types + [parfor_idx_t]
    name = "setidx_array%d_%s_par%d" % \
      (lhs_array_type.rank, lhs_array_type.elt_type, n_parfor_indices)
      
    #build_fn
    # Inputs:  
    #  - input_types
    #  - return_type=NoneType
    #  - name=None
    #  - input_names=None
    # Outputs:
    #  f, builder, input_vars 
    idx_names = ["idx%d" % (i+1) for i in xrange(n_fixed_indices)]
    start_names = ["start%d" % (i+1) for i in xrange(n_parfor_indices)]
    step_names = ["step%d" % (i+1) for i in xrange(n_parfor_indices)]
    
    input_names = ["output_array", "input_array"] + idx_names + start_names + step_names + ["sliceidx"]
    
    fn, builder, input_vars = build_fn(input_types, NoneType, name, input_names)
    lhs = input_vars[0]
    rhs = input_vars[1]
    fixed_indices = input_vars[2:(2+n_fixed_indices)]
    starts = input_vars[(2+n_fixed_indices):(2+n_fixed_indices+n_parfor_indices)]
    steps = input_vars[(2+n_fixed_indices+n_parfor_indices):(2+n_fixed_indices+2*n_parfor_indices)]
    assert (2+n_fixed_indices+2*n_parfor_indices+1) == len(input_vars), \
      "Wrong number of vars: %s, expected %d but got %d" % \
      (input_vars, 2+n_fixed_indices+2*n_parfor_indices+1, len(input_vars))
    parfor_idx = input_vars[-1]
    if n_parfor_indices > 1:
      parfor_indices = builder.tuple_elts(parfor_idx)
    else:
      parfor_indices = [parfor_idx]
    
    indices = []   
    # interleave the fixed and slice indices in their appropriate order
    # This would be easier if I had OCaml cons lists!
    slice_counter = 0
    fixed_counter = 0
    for i in xrange(n_indices):
      if fixed_counter < len(fixed_positions)  and i == fixed_positions[fixed_counter]:
        indices.append(fixed_indices[fixed_counter])
        fixed_counter += 1 
      else:
        assert slice_counter < len(slice_positions)  and slice_positions[slice_counter] == i, \
          "Bad positions for indices, missing %d" % i  
        start = starts[slice_counter]
        step = steps[slice_counter]
        parfor_idx = parfor_indices[slice_counter]
        indices.append(builder.add(start, builder.mul(step, parfor_idx)))
        slice_counter += 1
    
    value = builder.index(rhs, parfor_indices)
    builder.setidx(lhs, builder.tuple(indices), value)
    builder.return_(none)
    self._setidx_cache[key] = fn
    return fn
    
    
  def dissect_index_expr(self, expr):
    """
    Split up an indexing expression into 
    fixed scalar indices and the start/stop/step of all slices
    """

    if isinstance(expr.index.type, TupleT):
      indices = self.tuple_elts(expr.index)
    else:
      indices = [expr.index]
    
    n_dims = expr.value.type.rank 
    n_indices = len(indices)
    assert n_dims >= n_indices, \
      "Not yet supported: more indices (%d) than dimensions (%d) in %s" % (n_indices, n_dims, expr) 
    if n_indices < n_dims:
      indices = indices + [slice_none] * (n_dims - n_indices)
    
    if all_scalars(indices):
      # if there aren't any slice expressions, don't bother with the rest of this function
      return indices, range(len(indices)), [], []
    
    shape = self.shape(expr.value)
    shape_elts = self.tuple_elts(shape)
    slices = []
    slice_positions = []
    scalar_indices = []
    scalar_index_positions = []
    
    for i, shape_elt in enumerate(shape_elts):
      idx = indices[i]
      t = idx.type
      if isinstance(t, ScalarT):
        scalar_indices.append(idx)
        scalar_index_positions.append(i)
      elif isinstance(t, NoneT):
        slices.append( (zero_i64, shape_elt, one_i64) )
        slice_positions.append(i)
      else:
        assert isinstance(t, SliceT), "Unexpected index type: %s in %s" % (t, expr) 
        start = zero_i64 if t.start_type == NoneType else self.attr(idx, 'start')
        stop = shape_elt if t.stop_type == NoneType else self.attr(idx, 'stop')
        step = one_i64 if t.step_type == NoneType else self.attr(idx, 'step')
        slices.append( (start, stop, step) )
        slice_positions.append(i)   
    return scalar_indices, scalar_index_positions, slices, slice_positions
    
  def transform_Index(self, expr):

    ndims = expr.value.type.rank
    if ndims == 1 and isinstance(expr.index.type, ScalarT):
      return expr
    elif isinstance(expr.index.type, TupleT) and \
        len(expr.index.type.elt_types) == ndims and \
        all(isinstance(t, ScalarT) for t in expr.index.type.elt_types):
      return expr 
    
    scalar_indices, scalar_index_positions, slices, slice_positions = \
      self.dissect_index_expr(expr)
    assert len(scalar_indices) == len(scalar_index_positions)
    assert len(slices) == len(slice_positions)
    if len(slices) == 0:
      return expr
    
    array = self.transform_expr(expr.value)
    data = self.attr(array, 'data')
    shape = self.shape(array)
    shape_elts = self.tuple_elts(shape)
    strides = self.strides(array)
     
    stride_elts = self.tuple_elts(strides)
    offset = self.attr(array, 'offset')
    new_shape = []
    new_strides = []
    n_indices = len(scalar_indices) + len(slices)
    fixed_count = 0
    slice_count = 0

    for i in xrange(n_indices):
      if fixed_count < len(scalar_index_positions) and scalar_index_positions[fixed_count] == i:
        idx = scalar_indices[fixed_count]
        fixed_count += 1
        offset = self.add(offset, self.mul(idx, stride_elts[i]), name = "offset")

      else:
        assert slice_positions[slice_count] == i
        (start, stop, step) = slices[slice_count]
        slice_count += 1
        dim_offset = self.mul(start, stride_elts[i], name = "dim_offset")
        offset = self.add(offset, dim_offset, "offset")
        span = self.sub(stop, start, name = "span")
        shape_elt = self.div_round_up(span, step, name = "new_shape")
        new_shape.append(shape_elt)
        stride_elt = self.mul(stride_elts[i], step, name = "new_stride")
        new_strides.append(stride_elt)
 
    size = self.prod(new_shape)
    new_rank = len(slices)
    assert len(new_shape) == new_rank
    assert len(new_strides) == new_rank

    new_array = self.array_view(data, self.tuple(new_shape), self.tuple(new_strides), offset, size)
    assert new_array.type.rank == new_rank
    #if len(stride_elts) > 1:
    #  print "STRIDES", stride_elts 
    #  print "EXPR", expr 
    #  print "NEW ARRAY", new_array
    #  print 
    return new_array
    
    
      
  def assign_index(self, lhs, rhs):
    if isinstance(lhs.index.type, ScalarT) and isinstance(rhs.type, ScalarT):
      self.assign(lhs,rhs)
      return 
    
    scalar_indices, scalar_index_positions, slices, slice_positions = \
      self.dissect_index_expr(lhs)
    assert len(scalar_indices) == len(scalar_index_positions)
    assert len(slices) == len(slice_positions)
    if len(slices) == 0:
      self.setidx(lhs.value, self.tuple(scalar_indices), rhs)
      return 
    

    # if we've gotten this far then there is a slice somewhere in the indexing 
    # expression, so we're going to turn the assignment into a setidx parfor 
    bounds = self.tuple([self.div(self.sub(stop, start), step)
                         for (start, stop, step) in slices])
    
    setidx_fn = self.make_setidx_fn(lhs.value.type, 
                                    rhs.type, 
                                    scalar_index_positions, 
                                    slice_positions)
    starts = [start for (start, _, _) in slices]
    steps = [step for (_, _, step) in slices]
    closure = self.closure(setidx_fn, 
                           [lhs.value, rhs] + scalar_indices + starts + steps)


    self.parfor(closure, bounds)
  
  def transform_TypedFn(self, expr):
    if self.fn.created_by is not None and isinstance (self.fn.created_by, Phase):
      return self.fn.created_by.apply(expr)
    import pipeline
    return pipeline.indexify.apply(expr)
  
  def transform_Assign(self, stmt):
    lhs_class = stmt.lhs.__class__
    rhs = self.transform_expr(stmt.rhs)
     
    if lhs_class is Tuple:
      for (i, _) in enumerate(stmt.lhs.type.elt_types):
        lhs_i = self.tuple_proj(stmt.lhs, i)
        rhs_i = self.tuple_proj(rhs, i)
        # TODO: make this recursive, otherwise nested
        # complex assignments won't get implemented
        assert lhs_i.__class__ not in (ArrayView, Tuple)
        if lhs_i.__class__ is Index:
          self.assign_index(lhs_i)
        else:
          assert lhs_i.__class__ is Var, "Unexpcted LHS %s : %s" % (lhs_i, lhs_i.type)
          self.assign(lhs_i, rhs_i)
      return None
    elif lhs_class is Index:
      
      self.assign_index(stmt.lhs, rhs)
      return None
    else:
      stmt.rhs = rhs
      return stmt
    






































from .. import names 

class VarMap:
  def __init__(self):
    self._vars = {}

  def rename(self, old_name):
    new_name = names.refresh(old_name)
    self._vars[old_name] = new_name
    return new_name

  def lookup(self, old_name):
    if old_name in self._vars:
      return self._vars[old_name]
    else:
      return self.rename(old_name)

  def __str__(self):
    return "VarMap(%s)" % self._vars




from parakeet.testing_helpers import run_local_tests, expect, expect_eq

def rad2deg(rad):
  return (rad * 180) / 3.14159265359

import numpy as np 

def test_rad2deg_int32():
  rads = np.array([0,1]).astype('int32')
  degs = rad2deg(rads)
  expected = np.rad2deg(rads)
  expect_eq(degs, expected)

def test_rad2deg_bool():
  rads = np.array([False,True])
  degs = rad2deg(rads)
  expected = np.rad2deg(rads)
  expect_eq(degs, expected)

def test_rad2deg_float32():
  rads = np.array([0.1, 0.2]).astype('float32')
  degs = rad2deg(rads)
  expected = np.rad2deg(rads)
  expect_eq(degs, expected)


def test_rad2deg_float64():
  rads = np.array([0.1, 0.2]).astype('float64')
  degs = rad2deg(rads)
  expected = np.rad2deg(rads)
  expect_eq(degs, expected)

if __name__ == '__main__':
  run_local_tests()







"""
Authors: Federico Vaggi
License: MIT
Source: https://bitbucket.org/FedericoV/numpy-tip-complex-modeling/

Source: https://github.com/numfocus/python-benchmarks/blob/master/arc_distance/arc_distance_python.py
"""

from parakeet import jit, testing_helpers
import numpy as np
from math import * 
n = 10 

a = np.random.rand(n, 2)
b = np.random.rand(n, 2)


def arc_distance_for_loops(a, b):
    """
    Calculates the pairwise arc distance between all points in vector a and b.
    """
    a_nrows = a.shape[0]
    b_nrows = b.shape[0]

    distance_matrix = np.zeros([a_nrows, b_nrows])

    for i in range(a_nrows):
        theta_1 = a[i, 0]
        phi_1 = a[i, 1]
        for j in range(b_nrows):
            theta_2 = b[j, 0]
            phi_2 = b[j, 1]
            temp = (pow(sin((theta_2 - theta_1) / 2), 2)
                    +
                    cos(theta_1) * cos(theta_2)
                    * pow(sin((phi_2 - phi_1) / 2), 2))
            distance_matrix[i, j] = 2 * (atan2(sqrt(temp), sqrt(1 - temp)))
    return distance_matrix

def test_arc_distance_for_loops():
  testing_helpers.expect(arc_distance_for_loops, [a,b], arc_distance_for_loops(a,b))

"""
def arc_distance_tile(a, b):
    "Calculates the pairwise arc distance between all points in vector a and b."
    theta_1 = np.tile(a[:, 0], (b.shape[0], 1)).T
    phi_1 = np.tile(a[:, 1], (b.shape[0], 1)).T

    theta_2 = np.tile(b[:, 0], (a.shape[0], 1))
    phi_2 = np.tile(b[:, 1], (a.shape[0], 1))

    temp = (np.sin((theta_2 - theta_1) / 2)**2
            +
            np.cos(theta_1) * np.cos(theta_2)
            * np.sin((phi_2 - phi_1) / 2)**2)
    distance_matrix = 2 * (np.arctan2(np.sqrt(temp), np.sqrt(1 - temp)))

    return distance_matrix

def test_arc_distance_tile():
  testing_helpers.expect(arc_distance_tile, [a,b], arc_distance_tile(a,b))

"""

"""
# data-dependent broadcasting will never work with static
# map insertion so might as well drop this test since 
# it's hopeless
def arc_distance_broadcast(a, b):
    "Calculates the pairwise arc distance between all points in vector a and b."
    theta_1 = a[:, 0][:, None]
    theta_2 = b[:, 0][None, :]
    phi_1 = a[:, 1][:, None]
    phi_2 = b[:, 1][None, :]

    temp = (np.sin((theta_2 - theta_1) / 2)**2
            +
            np.cos(theta_1) * np.cos(theta_2)
            * np.sin((phi_2 - phi_1) / 2)**2)
    distance_matrix = 2 * (np.arctan2(np.sqrt(temp), np.sqrt(1 - temp)))
    return distance_matrix

def test_arc_distance_broadcast():
   testing_helpers.expect(arc_distance_broadcast, [a,b], arc_distance_broadcast(a,b))
"""

if __name__ == "__main__":
  testing_helpers.run_local_tests()
  



from dsltools import Node 

from .. ndtypes import Float32, Float64, Int32, Int64
from .. syntax import Var, Range, ConstArray, ConstArrayLike, IndexMap
from collect_vars import collect_var_names 
from syntax_visitor import SyntaxVisitor 


class AbstractValue(Node):
  _members = []

class ConstValue(AbstractValue):
  """
  All elements are the same constant 
  """
  _members = ['value', 'type']

zeros_float32 = ConstValue(0.0, Float32)
zeros_float64 = ConstValue(0.0, Float64)
zeros_int32 = ConstValue(0, Int32)
zeros_int64 = ConstValue(0, Int64)


class ConstElts(AbstractValue):
  """
  All elements known, represented as a numpy array 
  """ 
  _members = ['array']

class RangeArray(AbstractValue):
  """
  Result of a range expression 
  """
  _members = ['start', 'step', 'type']

class IndexMapResult(AbstractValue):
  _members = ['fn']

class Unknown(AbstractValue):
  _members = []
  
unknown = Unknown()

class IndexElimAnalysis(SyntaxVisitor):
    def __init__(self):
      # map variables to (start, stop, step)  
      self.index_vars = {}
      
      # map array names to abstract values 
      self.array_values = {}
      
      # map each array name to a set of names which share the same 
      # memory 
      self.linked_arrays = {}
      
    def visit_Assign(self, stmt):
      lhs = stmt.lhs
      rhs = stmt.rhs 
      
      if lhs.__class__ is Var:
        lhs_name = stmt.lhs.name
        
        rhs_class = rhs.__class__ 
        if rhs_class is Var:
          linked_set = self.linked_arrays.get(lhs_name, set([lhs_name]))
          rhs_name = stmt.rhs.name
          linked_set.add(rhs_name)
          self.linked_arrays[lhs_name] = linked_set 
          self.linked_arrays[rhs_name] = linked_set
          if lhs_name in self.array_values:
            self.array_values[rhs_name] = self.array_values[lhs_name]
        elif rhs_class is Range:
          self.array_values[lhs_name] = RangeArray(rhs.start, rhs.step, rhs.type.elt_type)
        elif rhs_class in (ConstArray, ConstArrayLike):
          self.array_values[lhs_name] = ConstValue(value = rhs.value, type = rhs.type) 
        elif rhs_class is IndexMap:
          self.array_values[lhs_name] = IndexMapResult(fn = rhs.fn)
      else:
        # if not a var, might be an index expression 
        for tainted_lhs_name in collect_var_names(lhs):
          for linked_name in self.linked_arrays.get(tainted_lhs_name, set([tainted_lhs_name])):
            self.array_values[linked_name] = unknown 


##########################
#  Performance Options   #
##########################
fast_math = True 
sse2 = True 
opt_level = '-O2'
# overload the default compiler path  
compiler_path = None

##########################
# Insert Debugging Code  #
##########################
debug = False
check_pyobj_types = False 

#########################
#  Verbose Printing     #
#########################
print_input_ir = False

print_line_numbers = False
 
print_function_source = False

print_commands = False
print_command_elapsed_time = False


# Generate a .c file or a .cpp? 
pure_c = True
delete_temp_files = True 

# if compiling C or OpenMP we can skip some of the craziness and 
# have distutils figure out the system config and compiler for us 
use_distutils = True

# show all the warnings? 
suppress_compiler_output = False 

# when compiling with NVCC, other headers get implicitly included 
# and cause warnings since Python redefines _POSIX_C_SOURCE  
# we can undefine it before including Python.h to get rid of those warnings 
undef_posix_c_source = True 


from parakeet.testing_helpers import expect, run_local_tests 
import numpy as np 

def always_1():
  return 1

def test_always_1():
  expect(always_1, [], 1)

def always_neg10():
  return -10

def test_always_neg10():
  expect(always_neg10, [], -10)

def always_true():
  return True

def test_always_true():
  expect(always_true, [], True, valid_types = (np.bool, np.bool8, np.bool_, bool))

def always_false():
  return False 

def test_always_false():
  expect(always_false, [], False, valid_types = (np.bool, np.bool8, np.bool_, bool))
  
def add1(x):
  return x + 1

def test_add1():
  expect(add1, [1],  2)

def call_add1(x):
  return add1(x)

def test_call_add1():
  expect(call_add1, [1],  2)

def call_nested_ident(x):
  def ident(y):
    return y
  return ident(x)

def test_nested_ident():
  expect(call_nested_ident, [1], 1)

global_val = 5
def use_global(x):
  return x + global_val

def test_use_global():
  expect(use_global, [3], 8)

def use_if_exp(x):
  return 1 if x < 10 else 2

def test_if_exp():
  expect(use_if_exp, [9], 1)
  expect(use_if_exp, [10], 2)

def simple_branch(x):
  if x < 10:
    return 1
  else:
    return 2

def test_simple_branch():
  expect(simple_branch, [9], 1)
  expect(simple_branch, [10], 2)

def simple_merge(x):
  if x == 0:
    y = 1
  else:
    y = x
  return y

def test_simple_merge():
  expect(simple_merge, [2], 2)
  expect(simple_merge, [0], 1)

def one_sided_merge(x,b):
  if b:
      x = 1
  return x

def test_one_sided_merge():
  expect(one_sided_merge, [100,True], 1)
  expect(one_sided_merge, [100,False], 100)

def if_true_const():
  if True:
    return 1
  else:
    return 2

def test_if_true_const():
  expect(if_true_const, [], 1)




if __name__ == '__main__':
  run_local_tests()




from dsltools import ScopedDict

from .. import prims, syntax 
from .. analysis.collect_vars import collect_var_names
from .. analysis.mutability_analysis import TypeBasedMutabilityAnalysis
from .. analysis.use_analysis import use_count
from .. ndtypes import ArrayT,  ClosureT, NoneT, ScalarT, TupleT, ImmutableT, NoneType, SliceT, FnT

from .. syntax import (AllocArray, Assign, ExprStmt, Expr, 
                       Const, Var, Tuple, TupleProj, Closure, ClosureElt, Cast,
                       Slice, Index, Array, ArrayView, Attribute, Struct, Select, 
                       PrimCall, Call, TypedFn, UntypedFn, 
                       OuterMap, Map, Reduce, Scan, IndexMap, IndexReduce, 
                       IndexScan, FilterReduce)
from .. syntax.helpers import (collect_constants, is_one, is_zero, is_false, is_true, all_constants,
                               get_types, 
                               slice_none_t, const_int, one, none, true, false, slice_none, 
                               zero_i64) 
import subst
import transform 
from transform import Transform


# classes of expressions known to have no side effects
# and to be unaffected by changes in mutable state as long
# as all their arguments are SSA variables or constants
#
# examples of unsafe expressions:
#  - Index: the underlying array is mutable, thus the expression depends on
#    any data modifications
#  - Call: Unless the function is known to contain only safe expressions it
#    might depend on mutable state or modify it itself

class Simplify(Transform):
  def __init__(self):
    transform.Transform.__init__(self)
    # associate var names with any immutable values
    # they are bound to
    self.bindings = ScopedDict()

    # which expressions have already been computed
    # and stored in some variable?
    self.available_expressions = ScopedDict()

  def pre_apply(self, fn):
    ma = TypeBasedMutabilityAnalysis()

    # which types have elements that might
    # change between two accesses?
    self.mutable_types = ma.visit_fn(fn)
    self.use_counts = use_count(fn)
    
  def immutable_type(self, t):
    return t not in self.mutable_types


  _immutable_classes = set([Const,  Var, 
                            Closure, ClosureElt, 
                            Tuple, TupleProj, 
                            Cast, PrimCall, 
                            TypedFn, UntypedFn, 
                            ArrayView, 
                            Slice, 
                            Map, Reduce, Scan, OuterMap, 
                            IndexMap, IndexReduce, IndexScan, 
                            ])
  
  def immutable(self, expr):
    """
    TODO: make all this mutability/immutability stuff sane 
    """
    klass = expr.__class__ 
    
    
    result = (klass in self._immutable_classes and 
                (all(self.immutable(c) for c in expr.children()))) or \
             (klass is Attribute and isinstance(expr.type, ImmutableT))
    return result 
    
                      
  def temp(self, expr, name = None, use_count = 1):
    """
    Wrapper around Codegen.assign_name which also updates bindings and
    use_counts
    """
    if self.is_simple(expr):
      return expr 
    else:
      new_var = self.assign_name(expr, name = name)
      self.bindings[new_var.name] = expr
      self.use_counts[new_var.name] = use_count
      return new_var

  def transform_expr(self, expr):
    if self.is_simple(expr):
      if expr.type == NoneType:
        return none 
      else:
        return Transform.transform_expr(self, expr)
    stored = self.available_expressions.get(expr)
    if stored is not None: 
      return stored
    return Transform.transform_expr(self, expr)

    
  def transform_Var(self, expr):
    t = expr.type 
    if t.__class__ is NoneT:
      return none 
    elif t.__class__ is SliceT and \
         t.start_type == NoneType and \
         t.stop_type == NoneType and \
         t.step_type == NoneType:
      return slice_none 
    
    name = expr.name
    prev_expr = expr

    while name in self.bindings:
      prev_expr = expr 
        
      expr = self.bindings[name]
      if expr.__class__ is Var:
        name = expr.name
      else:
        break
    c = expr.__class__ 
    
    if c is Var or c is Const:

      return expr
    else:

      return prev_expr
  
  def transform_Cast(self, expr):
    
    v = self.transform_expr(expr.value)
    if v.type == expr.type:
      return v
    elif v.__class__ is Const and isinstance(expr.type, ScalarT):
      return Const(expr.type.dtype.type(v.value), type = expr.type)
    elif self.is_simple(v):
      expr.value = v
      return expr
    else:
      expr.value = self.assign_name(v)
      return expr

  def transform_Attribute(self, expr):
    v = self.transform_expr(expr.value)
    
    if v.__class__ is Var and v.name in self.bindings:
      stored_v = self.bindings[v.name]
      c = stored_v.__class__
      if c is Var or c is Struct:
        v = stored_v
      elif c is ArrayView:
        if expr.name == 'shape':
          return self.transform_expr(stored_v.shape)
        elif expr.name == 'strides':
          return self.transform_expr(stored_v.strides)
        elif expr.name == 'data':
          return self.transform_expr(stored_v.data)
      elif c is AllocArray:
        if expr.name == 'shape':
          return self.transform_expr(stored_v.shape)
      elif c is Slice:
        if expr.name == "start":
          return self.transform_expr(stored_v.start)
        elif expr.name == "stop":
          return self.transform_expr(stored_v.stop)
        else:
          assert expr.name == "step", "Unexpected attribute for slice: %s" % expr.name  
          return self.transform_expr(stored_v.step)
    if v.__class__ is Struct:
      idx = v.type.field_pos(expr.name)
      return v.args[idx]
    elif v.__class__ is not Var:
      v = self.temp(v, "struct")
    if expr.value == v:
      return expr
    else:
      return Attribute(value = v, name = expr.name, type = expr.type)
  
  def transform_Closure(self, expr):
    expr.args = tuple(self.transform_simple_exprs(expr.args))
    return expr

  def transform_Tuple(self, expr):
    expr.elts = tuple( self.transform_simple_exprs(expr.elts))
    return expr

  def transform_TupleProj(self, expr):
    idx = expr.index
    assert isinstance(idx, int), \
        "TupleProj index must be an integer, got: " + str(idx)
    new_tuple = self.transform_expr(expr.tuple)

    if new_tuple.__class__ is Var and new_tuple.name in self.bindings:
      tuple_expr = self.bindings[new_tuple.name]
      
      if tuple_expr.__class__ is Tuple:
        assert idx < len(tuple_expr.elts), \
          "Too few elements in tuple %s : %s, elts = %s" % (expr, tuple_expr.type, tuple_expr.elts)
        return tuple_expr.elts[idx]
      elif tuple_expr.__class__ is Struct:
        assert idx < len(tuple_expr.args), \
          "Too few args in closure %s : %s, elts = %s" % (expr, tuple_expr.type, tuple_expr.elts) 
        return tuple_expr.args[idx]
    

    #if not self.is_simple(new_tuple):
    #  complex_expr = new_tuple 
    #  new_tuple = self.assign_name(complex_expr, "tuple")
    #  print "MADE ME A NEW TUPLE", complex_expr, new_tuple 
    expr.tuple = new_tuple
    return expr

  def transform_ClosureElt(self, expr):
    idx = expr.index
    assert isinstance(idx, int), \
        "ClosureElt index must be an integer, got: " + str(idx)
    new_closure = self.transform_expr(expr.closure)

    if new_closure.__class__ is Var and new_closure.name in self.bindings:
      closure_expr = self.bindings[new_closure.name]
      if closure_expr.__class__ is Closure:
        return closure_expr.args[idx]

    if not self.is_simple(new_closure):
      new_closure = self.assign_name(new_closure, "closure")
    expr.closure = new_closure
    return expr

  def transform_Call(self, expr):
    fn = self.transform_expr(expr.fn)
    args = self.transform_simple_exprs(expr.args)
    if fn.type.__class__ is ClosureT:
      closure_elts = self.closure_elts(fn)
      combined_args = tuple(closure_elts) + tuple(args)
      if fn.type.fn.__class__ is TypedFn:
        fn = fn.type.fn
      else:
        assert isinstance(fn.type.fn, UntypedFn)
        from .. type_inference import specialize 
        fn = specialize(fn, get_types(combined_args))
      assert fn.return_type == expr.type
      return Call(fn, combined_args, type = fn.return_type)
    else:
      expr.fn = fn
      expr.args = args
      return expr

  def transform_if_simple_expr(self, expr):
    if isinstance(expr, Expr):
      return self.transform_simple_expr(expr)
    else:
      return expr 
    
  def transform_simple_expr(self, expr, name = None):
    if name is None: name = "temp"
    result = self.transform_expr(expr)
    if not self.is_simple(result):
      return self.assign_name(result, name)
    else:
      return result
  
  def transform_simple_exprs(self, args):
    return [self.transform_simple_expr(x) for x in args]

  def transform_Array(self, expr):
    expr.elts = tuple(self.transform_simple_exprs(expr.elts))
    return expr
  
  def transform_Slice(self, expr):
    expr.start = self.transform_simple_expr(expr.start)
    expr.stop = self.transform_simple_expr(expr.stop)
    expr.step = self.transform_simple_expr(expr.step)
    return expr 

  
  def transform_index_expr(self, expr):
    if expr.__class__ is Tuple:
      new_elts = []
      for elt in expr.elts:
        new_elt = self.transform_expr(elt)
        if not self.is_simple(new_elt) and new_elt.type.__class__ is not SliceT:
          new_elt = self.temp(new_elt, "index_tuple_elt")
        new_elts.append(new_elt)
      expr.elts = tuple(new_elts)
      return expr 
    else:
      return self.transform_expr(expr) 
  
  def transform_Index(self, expr):
    expr.value = self.transform_expr(expr.value)

    expr.index = self.transform_index_expr(expr.index)

    if expr.value.__class__ is Array and expr.index.__class__ is Const:
      assert isinstance(expr.index.value, (int, long)) and \
             len(expr.value.elts) > expr.index.value
      return expr.value.elts[expr.index.value]
    
    # take expressions like "a[i][j]" and turn them into "a[i,j]" 
    if expr.value.__class__ is Index: 
      base_array = expr.value.value
      if isinstance(base_array.type, ArrayT):
        base_index = expr.value.index 
        if isinstance(base_index.type, TupleT):
          indices = self.tuple_elts(base_index)
        else:
          assert isinstance(base_index.type, ScalarT), \
            "Unexpected index type %s : %s in %s" % (base_index, base_index.type, expr)
          indices = [base_index]
        if isinstance(expr.index.type, TupleT):
          indices = tuple(indices) + tuple(self.tuple_elts(expr.index))
        else:
          assert isinstance(expr.index.type, ScalarT), \
            "Unexpected index type %s : %s in %s" % (expr.index, expr.index.type, expr)
          indices = tuple(indices) + (expr.index,)
        expr = self.index(base_array, self.tuple(indices))
        return self.transform_expr(expr)
    if expr.value.__class__ is not Var:
      expr.value = self.temp(expr.value, "array")
    return expr

  def transform_Struct(self, expr):
    new_args = self.transform_simple_exprs(expr.args)
    return syntax.Struct(new_args, type = expr.type)

  def transform_Select(self, expr):
    cond = self.transform_expr(expr.cond)
    trueval = self.transform_expr(expr.true_value)
    falseval = self.transform_expr(expr.false_value)
    if is_true(cond):
      return trueval 
    elif is_false(cond):
      return falseval
    elif trueval == falseval:
      return trueval 
    else:
      expr.cond = cond 
      expr.false_value = falseval 
      expr.true_value = trueval 
      return expr    
  
  def transform_PrimCall(self, expr):
    args = self.transform_simple_exprs(expr.args)
    prim = expr.prim
    if all_constants(args):
      return syntax.Const(value = prim.fn(*collect_constants(args)),
                          type = expr.type)
    
    if len(args) == 1:
      x = args[0]
      if prim == prims.logical_not:
        if is_false(x):
          return true 
        elif is_true(x):
          return false 
    if len(args) == 2:
      x,y = args 
      
      if prim == prims.add:
        if is_zero(x):
          return y
        elif is_zero(y):
          return x
        if y.__class__ is Const and y.value < 0:
          expr.prim = prims.subtract
          expr.args = (x, Const(value = -y.value, type = y.type))
          
          return expr 
        elif x.__class__ is Const and x.value < 0:
          expr.prim = prims.subtract
          expr.args = (y, Const(value = -x.value, type = x.type)) 
          return expr 
        
      elif prim == prims.subtract:
        if is_zero(y):
          return x
        elif is_zero(x) and y.__class__ is Var:
           
          stored = self.bindings.get(y.name)
          
          # 0 - (a * b) --> -a * b |or| a * -b 
          if stored and stored.__class__ is PrimCall and stored.prim == prims.multiply:
            
            a,b = stored.args
            if a.__class__ is Const:
              expr.prim = prims.multiply
              neg_a = Const(value = -a.value, type = a.type)
              expr.args = [neg_a, b]
              return expr 
            elif b.__class__ is Const:
              expr.prim = prims.multiply
              neg_b = Const(value = -b.value, type = b.type)
              expr.args = [a, neg_b]
              return expr 
            
      elif prim == prims.multiply:
      
        if is_one(x):
          return y
        elif is_one(y):
          return x
        elif is_zero(x):
          return x
        elif is_zero(y):
          return y
        
      elif prim == prims.divide and is_one(y):
        return x
      
      elif prim == prims.power:
      
        if is_one(y):
          return self.cast(x, expr.type)
        elif is_zero(y):
          return one(expr.type)
        elif y.__class__ is Const and y.value == 2:
          return self.cast(self.mul(x, x, "sqr"), expr.type)
      elif prim == prims.logical_and:
        if is_true(x):
          return y
        elif is_true(y):
          return x  
        elif is_false(x) or is_false(y):
          return false 
      elif prim == prims.logical_or:
        if is_true(x) or is_true(y):
          return true
        elif is_false(x):
          return y 
        elif is_false(y):
          return x 
    expr.args = args
    return expr 
  
  def transform_Map(self, expr):

    expr.args = self.transform_simple_exprs(expr.args)
    expr.fn = self.transform_expr(expr.fn)
    expr.axis = self.transform_if_expr(expr.axis)
    
    
    max_rank = max(self.rank(arg) for arg in expr.args)
    # if an axis is the Python value None, turn it into the IR expression for None
    if max_rank == 1 and self.is_none(expr.axis): expr.axis = zero_i64
    elif expr.axis is None: expr.axis = none   
    return expr  
  
  def transform_OuterMap(self, expr):
    expr.args = self.transform_simple_exprs(expr.args)
    expr.fn = self.transform_expr(expr.fn)
    expr.axis = self.transform_if_expr(expr.axis)
    max_rank = max(self.rank(arg) for arg in expr.args)
    # if an axis is the Python value None, turn it into the IR expression for None
    if max_rank == 1 and self.is_none(expr.axis): expr.axis = zero_i64
    elif expr.axis is None: expr.axis = none   
    return expr  
    
  def transform_shape(self, expr):
    if isinstance(expr, Tuple):
      expr.elts = tuple(self.transform_simple_exprs(expr.elts))
      return expr 
    else:
      return self.transform_simple_expr(expr)
  
  def transform_ParFor(self, stmt):
    stmt.bounds = self.transform_shape(stmt.bounds)
    stmt.fn = self.transform_expr(stmt.fn)
    return stmt
  
  def transform_Reduce(self, expr):
    expr.axis = self.transform_if_expr(expr.axis)
    expr.fn = self.transform_expr(expr.fn)
    expr.combine = self.transform_expr(expr.combine)
    expr.init = self.transform_if_simple_expr(expr.init)
    expr.args = self.transform_simple_exprs(expr.args)
    # if an axis is the Python value None, turn it into the IR expression for None
    max_rank = max(self.rank(arg) for arg in expr.args)
    if max_rank == 1 and self.is_none(expr.axis): expr.axis = zero_i64
    elif expr.axis is None: expr.axis = none   
    return expr  
  
  def transform_Scan(self, expr):
    expr.axis = self.transform_if_expr(expr.axis)
    expr.fn = self.transform_expr(expr.fn)
    expr.combine = self.transform_expr(expr.combine)
    expr.emit = self.transform_expr(expr.emit)
    expr.init = self.transform_if_simple_expr(expr.init)
    expr.args = self.transform_simple_exprs(expr.args)
    max_rank = max(self.rank(arg) for arg in expr.args)
    if max_rank == 1 and self.is_none(expr.axis): expr.axis = zero_i64
    elif expr.axis is None: expr.axis = none   
    return expr  
   
  
  def transform_IndexMap(self, expr):
    expr.fn = self.transform_expr(expr.fn)
    expr.shape = self.transform_shape(expr.shape)
    return expr 
  
  def transform_IndexReduce(self, expr):
    expr.fn = self.transform_if_expr(expr.fn)
    expr.combine = self.transform_expr(expr.combine)
    expr.init = self.transform_if_simple_expr(expr.init)
    expr.shape = self.transform_shape(expr.shape)
    return expr 
  
  def transform_IndexScan(self, expr):
    expr.fn = self.transform_if_expr(expr.fn)
    expr.combine = self.transform_expr(expr.combine)
    expr.emit = self.transform_if_expr(expr.emit)
    expr.init = self.transform_if_simple_expr(expr.init)
    expr.shape = self.transform_shape(expr.shape)
    return expr 
  
     
  def transform_ConstArray(self, expr):
    expr.shape = self.transform_shape(expr.shape)
    expr.value = self.transform_simple_expr(expr.value)
    return expr
  
  def transform_ConstArrayLike(self, expr):
    expr.array = self.transform_simple_expr(expr.array)
    expr.value = self.transform_simple_expr(expr.value)
  
  def temp_in_block(self, expr, block, name = None):
    """
    If we need a temporary variable not in the current top scope but in a
    particular block, then use this function. (this function also modifies the
    bindings dictionary)
    """
    if name is None:
      name = "temp"
    var = self.fresh_var(expr.type, name)
    block.append(Assign(var, expr))
    self.bindings[var.name] = expr
    return var

  def set_binding(self, name, value):
    assert value.__class__ is not Var or \
        value.name != name, \
        "Can't set name %s bound to itself" % name
    self.bindings[name] = value

  def bind_var(self, name, rhs):
    if rhs.__class__ is Var:
      old_val = self.bindings.get(rhs.name)
      if old_val and self.is_simple(old_val):
        self.set_binding(name, old_val)
      else:
        self.set_binding(name, rhs)
    else:
      self.set_binding(name, rhs)

  def bind(self, lhs, rhs):
    lhs_class = lhs.__class__
    if lhs_class is Var:
      self.bind_var(lhs.name, rhs)
    elif lhs_class is Tuple and rhs.__class__ is Tuple:
      assert len(lhs.elts) == len(rhs.elts)
      for lhs_elt, rhs_elt in zip(lhs.elts, rhs.elts):
        self.bind(lhs_elt, rhs_elt)

  def transform_lhs_Index(self, lhs):
    lhs.index = self.transform_index_expr(lhs.index)
    if lhs.value.__class__ is Var:
      stored = self.bindings.get(lhs.value.name)
      if stored and stored.__class__ is Var:
        lhs.value = stored
    else:
      lhs.value = self.assign_name(lhs.value, "array")
    return lhs

  def transform_lhs_Attribute(self, lhs):
    # lhs.value = self.transform_expr(lhs.value)
    return lhs

  def transform_ExprStmt(self, stmt):
    """Don't run an expression unless it possibly has a side effect"""

    v = self.transform_expr(stmt.value)
    if self.immutable(v):
      return None
    else:
      stmt.value = v
      return stmt

  def transform_Assign(self, stmt):
    
    lhs = stmt.lhs
    rhs = self.transform_expr(stmt.rhs)
    lhs_class = lhs.__class__
    rhs_class = rhs.__class__

    if lhs_class is Var:
      if lhs.type.__class__ is NoneT and self.use_counts.get(lhs.name,0) == 0:
        return self.transform_stmt(ExprStmt(rhs))
      elif self.immutable(rhs):
        
        self.bind_var(lhs.name, rhs)
        if rhs_class is not Var and rhs_class is not Const:
          self.available_expressions.setdefault(rhs, lhs)
    elif lhs_class is Tuple:
      self.bind(lhs, rhs)

    elif lhs_class is Index:
      if rhs_class is Index and \
         lhs.value == rhs.value and \
         lhs.index == rhs.index:
        # kill effect-free writes like x[i] = x[i]
        return None
      elif rhs_class is Var and \
           lhs.value.__class__ is Var and \
           lhs.value.name == rhs.name and \
           lhs.index.type.__class__ is TupleT and \
           all(elt_t == slice_none_t for elt_t in lhs.index.type.elt_types):
        # also kill x[:] = x
        return None
      else:
        lhs = self.transform_lhs_Index(lhs)
        # when assigning x[j] = [1,2,3]
        # just rewrite it as a sequence of element assignments 
        # to avoid 
        if lhs.type.__class__ is ArrayT and \
           lhs.type.rank == 1 and \
           rhs.__class__ is Array:
          lhs_slice = self.assign_name(lhs, "lhs_slice")
          for (elt_idx, elt) in enumerate(rhs.elts):
            lhs_idx = self.index(lhs_slice, const_int(elt_idx), temp = False)
            self.assign(lhs_idx, elt)
          return None
        elif not self.is_simple(rhs):
          rhs = self.assign_name(rhs)
    else:
      assert lhs_class is Attribute
      assert False, "Considering making attributes immutable"
      lhs = self.transform_lhs_Attribute(lhs)

    if rhs_class is Var and \
       rhs.name in self.bindings and \
       self.use_counts.get(rhs.name, 1) == 1:
      self.use_counts[rhs.name] = 0
      rhs = self.bindings[rhs.name]
    stmt.lhs = lhs
    stmt.rhs = rhs
    return stmt

  def transform_block(self, stmts, keep_bindings = False):
    self.available_expressions.push()
    self.bindings.push()
    
    new_stmts = Transform.transform_block(self, stmts)
    
    self.available_expressions.pop()
    if not keep_bindings:
      self.bindings.pop()
    return new_stmts

  def enter_loop(self, phi_nodes):
    result = {}
    for (k, (left,right)) in phi_nodes.iteritems():
      new_left = self.transform_expr(left)
      if new_left == right:
        self.set_binding(k, new_left)
      else:
        result[k] = (new_left, right)
    return result 
  
  def transform_merge(self, phi_nodes, left_block, right_block):
    result = {}
    for (k, (left, right)) in phi_nodes.iteritems():
      new_left = self.transform_expr(left)
      new_right = self.transform_expr(right)

      if not isinstance(new_left, (Const, Var)):
        new_left = self.temp_in_block(new_left, left_block)
      if not isinstance(new_right, (Const, Var)):
        new_right = self.temp_in_block(new_right, right_block)

      if new_left == new_right:
        # if both control flows yield the same value then
        # we don't actually need the phi-bound variable, we can just
        # replace the left value everywhere
        self.assign(Var(name= k, type = new_left.type), new_left)
        self.set_binding(k, new_left)
      else:
        result[k] = new_left, new_right
    return result

  def transform_If(self, stmt):
    stmt.true = self.transform_block(stmt.true, keep_bindings = True)
    stmt.false = self.transform_block(stmt.false, keep_bindings=True)
    stmt.merge = self.transform_merge(stmt.merge,
                                      left_block = stmt.true,
                                      right_block = stmt.false)
    self.bindings.pop()
    self.bindings.pop()
    stmt.cond = self.transform_simple_expr(stmt.cond, "cond")
    if len(stmt.true) == 0 and len(stmt.false) == 0 and len(stmt.merge) <= 2:
      for (lhs_name, (true_expr, false_expr)) in stmt.merge.items():
        lhs_type = self.lookup_type(lhs_name)
        lhs_var = Var(name = lhs_name, type = lhs_type)
        assert true_expr.type == false_expr.type, \
          "Unexpcted type mismatch: %s != %s" % (true_expr.type, false_expr.type)
        rhs = Select(stmt.cond, true_expr, false_expr, type = true_expr.type)
        self.bind_var(lhs_name, rhs)
        self.assign(lhs_var, rhs)
      return None 
    return stmt

  def transform_loop_condition(self, expr, outer_block, loop_body, merge):
    """Normalize loop conditions so they are just simple variables"""

    if self.is_simple(expr):
      return self.transform_expr(expr)
    else:
      loop_carried_vars = [name for name in collect_var_names(expr)
                           if name in merge]
      if len(loop_carried_vars) == 0:
        return expr

      left_values = [merge[name][0] for name in loop_carried_vars]
      right_values = [merge[name][1] for name in loop_carried_vars]

      left_cond = subst.subst_expr(expr, dict(zip(loop_carried_vars,
                                                  left_values)))
      if not self.is_simple(left_cond):
        left_cond = self.temp_in_block(left_cond, outer_block, name = "cond")

      right_cond = subst.subst_expr(expr, dict(zip(loop_carried_vars,
                                                   right_values)))
      if not self.is_simple(right_cond):
        right_cond = self.temp_in_block(right_cond, loop_body, name = "cond")

      cond_var = self.fresh_var(left_cond.type, "cond")
      merge[cond_var.name] = (left_cond, right_cond)
      return cond_var

    
  def transform_While(self, stmt):
    merge = self.enter_loop(stmt.merge)
    stmt.body = self.transform_block(stmt.body)
    stmt.merge = self.transform_merge(merge,
                                      left_block = self.blocks.current(),
                                      right_block = stmt.body)
    stmt.cond = \
        self.transform_loop_condition(stmt.cond,
                                      outer_block = self.blocks.current(),
                                      loop_body = stmt.body,
                                      merge = stmt.merge)
    return stmt

  def transform_ForLoop(self, stmt):
    

    merge = self.enter_loop(stmt.merge)
    stmt.body = self.transform_block(stmt.body) 
    stmt.merge = self.transform_merge(merge,
                                      left_block = self.blocks.current(),
                                      right_block = stmt.body)
    stmt.start = self.transform_simple_expr(stmt.start, 'start')
    stmt.stop = self.transform_simple_expr(stmt.stop, 'stop')
    if self.is_none(stmt.step):
      stmt.step = one(stmt.start.type)
    else:
      stmt.step = self.transform_simple_expr(stmt.step, 'step')

    # if a loop is only going to run for one iteration, might as well get rid of
    # it
    if stmt.start.__class__ is Const and \
       stmt.stop.__class__ is Const and \
       stmt.step.__class__ is Const:
      if stmt.start.value >= stmt.stop.value:
        for (var_name, (input_value, _)) in stmt.merge.iteritems():
          var = Var(var_name, input_value.type)
          self.blocks.append(Assign(var, input_value))
        return None
      elif stmt.start.value + stmt.step.value >= stmt.stop.value:
        for (var_name, (input_value, _)) in stmt.merge.iteritems():
          var = Var(var_name, input_value.type)
          self.blocks.append(Assign(var, input_value))
        self.assign(stmt.var, stmt.start)
        self.blocks.top().extend(stmt.body)
        return None
    return stmt

  def transform_Return(self, stmt):
    new_value = self.transform_expr(stmt.value)
    if new_value != stmt.value:
      stmt.value = new_value
    return stmt




from .. import prims 
from ..frontend import jit, macro, typed_macro 
from ..ndtypes import ScalarT, ArrayT, make_array_type
from ..syntax import Select, DelayUntilTyped, PrimCall, Call,  OuterMap



 
@macro
def _select(cond, trueval, falseval):
  return Select(cond, trueval, falseval)

#@macro  
def where(cond, trueval, falseval):
  #return Select(cond, trueval, falseval )
  return map(_select, cond, trueval, falseval)


@jit 
def diff(x):
  """
  TODO:
    - axis selection
    - preserve size by filling with zeros
    - allow n'th differences by recursion
  """
  return x[1:] - x[:-1]


@jit 
def fill(x, v):
  for i in range(len(x)):
    x[i] = v 
 




from ..syntax import Assign, ExprStmt, If, While, ForLoop, Comment, Return, ParFor  

def all_branches_return(stmt):
  if isinstance(stmt, Return):
    return True 
  elif isinstance(stmt, If):
    return len(stmt.true) > 0 and \
      all_branches_return(stmt.true[-1]) and \
      len(stmt.false) > 0 and \
      all_branches_return(stmt.false[-1])

def can_inline_block(stmts, outer = False):
  for stmt in stmts:
    stmt_class = stmt.__class__
    if stmt_class in (Assign, ExprStmt, ParFor):
      pass
    elif stmt_class is If:
      # if every branch leading from here ends up 
      # returning a value, then it's OK to replace those
      # with variable assignments  
      # ...but only once the Inliner knows how to deal with 
      # control flow...      
      #if outer and all_branches_return(stmt):
      #  return True
      if not can_inline_block(stmt.true, outer = False):
        return False
      if not can_inline_block(stmt.false, outer=False):
        return False 
          
    elif stmt_class in (While, ForLoop):
      if not can_inline_block(stmt.body):
        return False
    elif stmt_class is Comment:
      continue
    
    else:
      assert stmt_class is Return, "Unexpected statement: %s" % stmt
      if not outer:
        return False 
  return True

def can_inline(fundef):
  return can_inline_block(fundef.body, outer = True)


import parakeet as par
from parakeet import testing_helpers, jit 
import numpy as np


class Network(object):
  def __init__(self):
    self.layers = []
    
    self.start_epoch()
    
  def start_epoch(self):
    self.sse = 0
    self.n_updates = 0
     
  def mse(self):
    return self.sse / self.n_updates 
   
  def __iadd__(self, layer):
    self.layers.append(layer)
    return self
    
  def predict(self, x):
    for layer in self.layers:
      x = layer.fprop(x)
    return x
  
  def bprop(self, err):
    for layer in reversed(self.layers[1:]):
      err = layer.bprop(err)
  
  def update(self, x, y):
    y_pred = self.predict(x)
    err = y_pred - y
    self.sse += err**2 
    self.n_updates += 1
    self.bprop(err)


@jit
def sigmoid(x):
  return 1.0 / (1.0 + par.exp(-x))

@jit
def d_sigmoid(x):
  s = sigmoid(x)
  return s * (1-s) 

@jit
def dot(x,y):
  return sum(x*y)

@jit
def fprop_linear(x, W, b):
  def dot_add(w_row, b_elt):
    return sum(w_row * x) + b_elt 
  return par.each(dot_add, W, b)
 
@jit 
def fprop_logistic(x, W, b):
  return sigmoid(fprop_linear(x,W,b))

@jit
def bprop_logistic(x, W, b, err):
  return W*0.001

class LogisticLayer(object):
  def __init__(self, n_in, n_out, learning_rate = 0.001):
    self.W = np.random.randn( n_out, n_in)
    self.bias = np.random.randn(n_out)
    self.last_input = None
    self.last_output = None 
    self.learning_rate = learning_rate
    
  def __str__(self):
    return "LogisticLayer(in = %d, out = %d)" % (self.W.shape[0], self.W.shape[1])
  
  def fprop(self, x):
    self.last_input = x
    self.last_output = fprop_logistic(x, self.W, self.bias)

    return self.last_output
  
  def bprop(self, err):
    weighted_err = 0  
    return weighted_err 
    

@jit 
def tanh_fprop(x, w, b):
  return par.tanh(sum(x*w) + b)

def tanh_bprop(x, w, b, err):
  return 0

class TanhLayer(object):
  def __init__(self, n, w = None, bias = None):
    if w is None:
      w = np.random.randn(n)
    if bias is None:
      bias = np.random.randn()
    self.w = w
    self.bias = bias
    self.last_x = None
    
  def fprop(self, data):
    self.last_data = data
    self.last_output = tanh_fprop(data, self.w, self.bias)
    return self.last_output
     
  def bprop(self, err):
    self.last_delta = tanh_bprop(self.last_data, self.w, self.bias, err)
    

n_in = 1000 
n_hidden = 50
n_out = 1 

mlp = Network()
mlp += LogisticLayer(n_in, n_hidden)
mlp += LogisticLayer(n_hidden, n_out)

def test_mlp():
  x = np.random.randn(1000)
  y = mlp.predict(x)
  assert len(y) == n_out
  assert np.all(y >= 0)
  assert np.all(y <= 1)
  


if __name__ == '__main__':
  testing_helpers.run_local_tests()



from .. import prims 
from .. analysis.offset_analysis import OffsetAnalysis
from .. syntax import Var, Const 
from .. syntax.helpers import true, false, const
from transform import Transform 


class OffsetPropagation(Transform):
  def pre_apply(self, fn):
    # map variables to pairs which include low/exclude high  
    self.known_ranges = {}
    self.known_offsets = OffsetAnalysis().visit_fn(fn)
    self.seen_binding = set(fn.arg_names)

  def transform_Assign(self, stmt):
    result = Transform.transform_Assign(self, stmt )
    if result.lhs.__class__ is Var:
      self.seen_binding.add(result.lhs.name)
    return result 
  
  def transform_ForLoop(self, stmt):
    self.known_ranges[stmt.var.name] = (stmt.start, stmt.stop)
    stmt.body = self.transform_block(stmt.body)
    return stmt 
  
  def compare_lt(self, x, y, inclusive = True):
    if x.__class__ is Var:
      offsets = [(x.name, 0)]
      if x.name in self.known_offsets:
        offsets = self.known_offsets[x.name].union(set(offsets))
      for (var_name, offset) in offsets:
        if var_name in self.known_ranges:
          (low,high) = self.known_ranges[var_name]
          if y == high:
            if (offset >= 0 and inclusive) or (offset > 0):
              return false
          elif y == low:
            if (offset <= 0 and inclusive) or (offset < 0):
              return true
    return None 
  
  def const_additive_cancellation(self, x_name, y_value, output_type):
    """
    Return a single variable in situations like 
      x = prev_var + const 
      y = x + (- const) 
    """ 
    if x_name in self.known_offsets:
      for (prev_var_name, offset) in self.known_offsets[x_name]:
        if prev_var_name in self.seen_binding:
          if offset == -y_value:
            prev_type = self.fn.type_env[prev_var_name]
            prev_var = Var(name=prev_var_name, type = prev_type)
            return self.cast(prev_var, output_type)
  
  def var_subtractive_cancellation(self, x_name, y_name, output_type):
    """
    Return a single constant in cases like:
      a = b + const 
      c = b - a
    """
    if x_name in self.known_offsets:
      for (prev_var_name, offset) in self.known_offsets[x_name]:
        if prev_var_name == y_name:
          return self.cast(const(offset), output_type)
        
  
  def transform_PrimCall(self, expr):
    p = expr.prim
    
    result = None 
    if isinstance(p, prims.Cmp):
      x,y = expr.args   
      if p == prims.less_equal:
        result = self.compare_lt(x,y, inclusive=True)
      elif p == prims.less:
        result = self.compare_lt(x,y, inclusive=False)
      elif p == prims.greater_equal:
        result = self.compare_lt(y, x, inclusive=True)
      elif p == prims.greater:
        result = self.compare_lt(y, x, inclusive=False)
      elif p == prims.equal:
        result1 = self.compare_lt(x, y, inclusive=True)
        result2 = self.compare_lt(y, x, inclusive=True)
        if result1 is not None and result2 is not None and result1 == result2:
          result = result1 
          
          
    # THIS IS MUCH MORE LIMITED THAN THE SORT OF SYMBOLIC REWRITES
    # WE COULD DO IN GENERAL, i.e. y = x + 2, z = x - 2, (y-z) ==> 4
    elif p == prims.add:
      x, y = expr.args
      result = None  
      if x.__class__ is Var and y.__class__ is Const:
        result = self.const_additive_cancellation(x.name, y.value, expr.type)
      elif x.__class__ is Const and y.__class__ is Var:
        result = self.const_additive_cancellation(y.name, x.value, expr.type)
        
      if result is not None: return result
         
    elif p == prims.subtract:

      x, y = expr.args 
      result = None 
      if x.__class__ is Var and y.__class__ is Const:
        result = self.const_additive_cancellation(x.name, -y.value, expr.type)
      elif x.__class__ is Var and y.__class__ is Var:
        result = self.var_subtractive_cancellation(x.name, y.name, expr.type)
      if result is not None: return result
    if result is not None:
      return result
    else:
      return expr
      



from parakeet import UInt8, Int8, Int32, Int64, Float32, Float64, Bool
from parakeet.testing_helpers import expect_type, run_local_tests

def add1(x):
  return x + 1

def call_add1(x):
  return add1(x)

def test_add1():
  expect_type(add1, [Int32], Int64)
  expect_type(add1, [Int8], Int64)
  expect_type(add1, [UInt8], Int64)
  expect_type(add1, [Bool], Int64)
  expect_type(add1, [Int64], Int64)
  expect_type(add1, [Float32], Float64)

def test_call_add1():
  expect_type(call_add1, [Int32], Int64)
  expect_type(call_add1, [Float32], Float64)
  expect_type(call_add1, [Bool], Int64)
  expect_type(add1, [Float32], Float64)

def add(x,y):
  return x + y

def test_add_bools():
  """
  test_add_bools:
  Parakeet booleans don't behave like the default Python type but rather like
  numpy.bool8, where (+) == (or) and (*) == (and)
  """
  expect_type(add, [Bool, Bool], Bool)

def branch_return(b):
  if b:
    return 1
  else:
    return 1.0

def test_branch_return():
  expect_type(branch_return, [Bool], Float64)

def branch_assign(b):
  if b:
    x = 0
  else:
    x = 1.0
  return x

def test_branch_assign():
  expect_type(branch_assign, [Bool], Float64)

def incr_loop(init, count):
  x = init
  while x < count:
    x = x + 1
  return x

def test_incr_loop():
  expect_type(incr_loop, [Int32, Int32], Int64)
  expect_type(incr_loop, [Float64, Int32], Float64)

if __name__ == '__main__':
  run_local_tests()



import numpy as np
import parakeet
from parakeet import jit 
from parakeet.testing_helpers import expect, run_local_tests, eq


@jit 
def fill_with_const(a, k):
  def return_const(idx):
    return k
  return parakeet.imap(return_const, a.shape)
  
def test_fill_with_const():
  shape=(1,2,1,1)
  a = np.empty(shape=shape, dtype='float32')
  k = 3.137
  expected = np.ones(shape = shape, dtype='float32') * k
  expect(fill_with_const, [a, k], expected)


def test_identity():
  x = np.ones((1,1,2,1))
  def get_idx(idx):
    return x[idx]
  y = parakeet.imap(get_idx, x.shape)
  assert eq(x,y)

if __name__ == '__main__':
  run_local_tests()



from .. syntax import Var, Tuple
from syntax_visitor import SyntaxVisitor

class SetCollector(SyntaxVisitor):
  def __init__(self):
    SyntaxVisitor.__init__(self)
    self.var_names = set([])

  def visit_Var(self, expr):
    self.var_names.add(expr.name)

def collect_var_names(expr):
  collector = SetCollector()
  collector.visit_expr(expr)
  return collector.var_names

def collect_var_names_from_exprs(exprs):
  collector = SetCollector()
  collector.visit_expr_list(exprs)
  return collector.var_names

class ListCollector(SyntaxVisitor):
  def __init__(self):
    SyntaxVisitor.__init__(self)
    self.var_names = []

  def visit_Var(self, expr):
    self.var_names.append(expr.name)

def collect_var_names_list(expr):
  collector = ListCollector()
  collector.visit_expr(expr)
  return collector.var_names



def collect_binding_names(lhs):
  lhs_class = lhs.__class__
  if lhs_class is Var:
    return [lhs.name]
  elif lhs.__class__ is Tuple:
    combined = []
    for elt in lhs.elts:
      combined.extend(collect_binding_names(elt))
    return combined
  else:
    return []

class CollectBindings(SyntaxVisitor):
  def __init__(self):
    SyntaxVisitor.__init__(self)
    self.bindings = {}

  def bind(self, lhs, rhs):
    if lhs.__class__ is Var:
      self.bindings[lhs.name] = rhs
    elif lhs.__class__ is Tuple:
      for elt in lhs.elts:
        self.bind(elt, rhs)

  def visit_Assign(self, stmt):
    self.bind(stmt.lhs, stmt.rhs)

def collect_bindings(fn):
  return CollectBindings().visit_fn(fn)



import numpy as np 

class AbstractValue(object):
  def __repr__(self):
    return str(self)
  
  def __hash__(self):
    assert False, "Hash function not implemented for %s : %s" % (self, self.__class__)
  
  def __eq__(self):
    return False 
  
  def __ne__(self, other):
    return not (self == other)

class Unknown(AbstractValue):
  def __str__(self):
    return "unknown"
  
  def __eq__(self, other):
    return other.__class__ is Unknown 
  
  def __hash__(self):
    return 107

unknown = Unknown()

class Tuple(AbstractValue):
  
  __slots__ = ['elts', '_hash']
  
  def __init__(self, elts):
    self.elts = tuple(elts)
    self._hash = hash(self.elts)
  
  def __str__(self):
    return "Tuple(%s)" % ", ".join(str(elt) for elt in self.elts)
  
  def __hash__(self):
    return self._hash 
  
  def __eq__(self, other):
    return other.__class__ is Tuple and self.elts == other.elts 
      
  
class Array(AbstractValue):
  # mark known strides with integer constants 
  # and all others as unknown
  
  __slots__ = ['strides', '_hash']
  
  def __init__(self, strides):
    self.strides = strides
    self._hash = hash(self.strides.elts) + 1
  
  def __str__(self):
    return "Array(strides = %s)" % self.strides
  
  def __hash__(self):
    return self._hash 
  
  def __eq__(self, other):
    return other.__class__ is Array and \
      self.strides == other.strides


class Struct(AbstractValue):
  def __init__(self, fields):
    self.fields = fields 
    
  def __str__(self):
    return "Struct(%s)" % self.fields
  
  def __eq__(self, other):
    if other.__class__ != Struct:
      return False
    my_fields = set(self.fields.keys())
    other_fields = set(other.fields.keys())
    if my_fields != other_fields:
      return False 
    for f in my_fields:
      if self.fields[f] != other.fields[f]:
        return False
    return True
  
  def __hash__(self):
    return hash(tuple(self.fields.items()))

   
class Const(AbstractValue):
  __slots__ = ['value']
  def __init__(self, value):
    self.value = value
  
  def __hash__(self):
    return int(self.value) 
  
  def __str__(self):
    return str(self.value)
  
  def __eq__(self, other):
    return other.__class__ is Const and self.value == other.value

zero = Const(0)
one = Const(1)

def specialization_const(x, specialize_all = False):
  if x == 0:
    return zero 
  elif x == 1:
    return one
  elif specialize_all:
    return Const(x)
  else:
    return unknown

def abstract_tuple(elts):
  return Tuple(tuple(elts))

def abstract_array(strides):
  return Array(abstract_tuple(strides))





from array_type import (ArrayT, make_array_type,
                        elt_type, elt_types, rank,
                        get_rank, lower_rank,
                        lower_rank, lower_ranks, increase_rank)

from closure_type import make_closure_type, ClosureT 

from core_types import (Type, AnyT, Any, UnknownT, Unknown, 
                        TypeFailure, IncompatibleTypes, FieldNotFound,
                        NoneT, NoneType, 
                        ConcreteT, ImmutableT, StructT, 
                        TypeValueT, combine_type_list)

from fn_type import make_fn_type, FnT 

from ptr_type import PtrT, ptr_type 

from scalar_types import (ScalarT, 
                          IntT, FloatT, BoolT, SignedT, UnsignedT,  
                          Int8, Int16, Int24, Int32, Int64, 
                          UInt8, UInt16, UInt32, UInt64, 
                          Bool, Float32, Float64, 
                          is_scalar_subtype, is_scalar, all_scalars, 
                          from_dtype, from_char_code)
  
from slice_type import SliceT, make_slice_type

from tuple_type import TupleT, make_tuple_type, empty_tuple_t, repeat_tuple

import dtypes
import type_conv   
from type_conv import typeof




import numpy as np
from numpy import  asarray 
import types 


from ..syntax import UntypedFn, TypedFn 
from ..ndtypes import (scalar_types, 
                       make_array_type, ArrayT, 
                       make_tuple_type, TupleT, 
                       make_closure_type, ClosureT, 
                       NoneT, NoneType, TypeValueT)
from ..ndtypes.type_conv import typeof, register   

register(type(None), NoneT, lambda _: NoneType)


def typeof_dtype(dt):
  return TypeValueT(scalar_types.from_dtype(dt))

register([np.dtype], TypeValueT, typeof_dtype) 

def typeof_type(t):
  assert hasattr(t, 'dtype'), "Can only convert numpy types"
  dt = t(0).dtype 
  pt = scalar_types.from_dtype(dt)
  return TypeValueT(pt) 

register(types.TypeType, TypeValueT, typeof_type)

def typeof_tuple(python_tuple):
  return make_tuple_type([typeof(elt_t) for elt_t in python_tuple])

register(types.TupleType, TupleT, typeof_tuple)

def typeof_array(x):
  x = asarray(x)
  elt_t = scalar_types.from_dtype(x.dtype)
  return make_array_type(elt_t, x.ndim)

register(np.ndarray, ArrayT, typeof_array)

from .. import prims 
register((list, xrange), ArrayT, typeof_array)



from decorators  import jit, macro 

 
# ndtypes already register NumPy arrays type converters, 
# but parakeet also treats ranges and lists as arrays 
register((list, xrange), ArrayT, typeof_array)

def typeof_prim(p):
  from ..syntax.wrappers import  build_untyped_prim_fn
  untyped_fn = build_untyped_prim_fn(p)
  return make_closure_type(untyped_fn, ())

register(prims.class_list, ClosureT, typeof_prim)

def typeof_fn(f):
  import ast_conversion
  untyped_fn = ast_conversion.translate_function_value(f)
  closure_args = untyped_fn.python_nonlocals()
  closure_arg_types = map(typeof, closure_args)
  return make_closure_type(untyped_fn, closure_arg_types)


register(types.FunctionType, ClosureT, typeof_fn)
register(jit, ClosureT, typeof_fn)
register(macro, ClosureT, typeof_fn)
register(types.BuiltinFunctionType, ClosureT, typeof_fn)
register(np.ufunc, ClosureT, typeof_fn)
register(UntypedFn, ClosureT, typeof_fn)
register(TypedFn, ClosureT, typeof_fn)



from parakeet import jit 
from parakeet.testing_helpers import eq, run_local_tests, expect 

from numpy import exp, log, sqrt


def CND(x):
  a1 = 0.31938153
  a2 = -0.356563782
  a3 = 1.781477937
  a4 = -1.821255978
  a5 = 1.330274429
  L = abs(x)
  K = 1.0 / (1.0 + 0.2316419 * L)
  w = 1.0 - 1.0/sqrt(2*3.141592653589793)* exp(-1*L*L/2.) * (a1*K +
      a2*K*K + a3*K*K*K + a4*K*K*K*K + a5*K*K*K*K*K)
  if x<0:
    w = 1.0-w
  return w

def black_scholes(CallFlag,S,X,T,r,v):
  d1 = ((r+v*v/2.)*T+log(S/X))/(v*sqrt(T))
  d2 = d1-v*sqrt(T)
  z = exp(-1.0*r*T) * X
  if CallFlag:
    return S*CND(d1) - z*CND(d2)
  else:
    return z*CND(-1.0*d2) - S*CND(-1.0*d1)

black_scholes_parakeet = jit(black_scholes)

def test_black_scholes():
  x1 = (False, 10.0, 10.0, 2.0, 2.0, 2.0)
  x2 = (True, 10.0, 10.0, 2.0, 2.0, 2.0)
  xs = [x1, x2]
  for x in xs:
    expect(black_scholes, x, black_scholes(*x))

if __name__ == '__main__':
  run_local_tests()
